import os
import platform
import re
import typing
from pathlib import Path

from ._api import OntologyStore, RemoteOntologyService, OntologyReleaseService
from ._github import GitHubRemoteOntologyService, GitHubOntologyReleaseService


def configure_ontology_store(
    store_dir: typing.Optional[str] = None,
    ontology_release_service: OntologyReleaseService = GitHubOntologyReleaseService(),
    remote_ontology_service: RemoteOntologyService = GitHubRemoteOntologyService(),
) -> OntologyStore:
    """
    Configure and create the default ontology store.

    :param store_dir: a `str` pointing to an existing directory for caching the ontology files
      or `None` if the platform-specific default folder should be used.
    :param ontology_release_service: an :class:`OntologyReleaseService` for fetching the ontology releases.
    :param remote_ontology_service: a :class:`RemoteOntologyService` responsible for fetching
      the ontology data from a remote location if we do not have the data locally.
    :returns: an :class:`OntologyStore`.
    :raises: `ValueError` if something goes wrong.
    """
    if store_dir is None:
        store_dir = get_default_ontology_store_dir()
    else:
        if not os.path.isdir(store_dir):
            raise ValueError("`store_dir` must point to an existing directory")
    return OntologyStore(
        store_dir=store_dir,
        ontology_release_service=ontology_release_service,
        remote_ontology_service=remote_ontology_service,
    )


def get_default_ontology_store_dir() -> str:
    """
    Get a platform specific absolute path to the data directory.

    The data directory points to `$HOME/.hpo-toolkit` on UNIX and `$HOME/hpo-toolkit` on Windows.
    The folder is created if it does not exist.
    """
    ps = platform.system()

    if re.match("(linux)|(darwin)", ps, re.IGNORECASE):
        store_dir_name = ".hpo-toolkit"
    elif re.match("windows", ps, re.IGNORECASE):
        store_dir_name = "hpo-toolkit"
    else:
        raise ValueError(f"Unsupported platform {ps}")

    dir_name = os.path.join(Path.home(), store_dir_name)

    if not os.path.isdir(dir_name):
        os.makedirs(dir_name)

    return dir_name



import pytest

from hpotk import TermId, OntologyGraph
from hpotk.graph import IncrementalCsrGraphFactory
from ._hierarchical import (
    HierarchicalIcTermIdSorting,
    HierarchicalEdgeTermIdSorting,
    EdgeSimilarityMeasure,
)


def make_nodes_and_edges():
    """Prepare nodes and edges for testing."""
    # Nodes are sorted.
    nodes = [
        TermId.from_curie("HP:01"),
        TermId.from_curie("HP:010"),
        TermId.from_curie("HP:011"),
        TermId.from_curie("HP:0110"),
        TermId.from_curie("HP:02"),
        TermId.from_curie("HP:020"),
        TermId.from_curie("HP:021"),
        TermId.from_curie("HP:022"),
        TermId.from_curie("HP:03"),
        TermId.from_curie("HP:1"),
    ]

    # Edges for a graph with a single root node: HP:1
    edges = [
        # source -> destination
        (TermId.from_curie("HP:01"), TermId.from_curie("HP:1")),
        (TermId.from_curie("HP:010"), TermId.from_curie("HP:01")),
        (TermId.from_curie("HP:011"), TermId.from_curie("HP:01")),
        # This one is multi-parent.
        (TermId.from_curie("HP:0110"), TermId.from_curie("HP:010")),
        (TermId.from_curie("HP:0110"), TermId.from_curie("HP:011")),
        (TermId.from_curie("HP:02"), TermId.from_curie("HP:1")),
        (TermId.from_curie("HP:020"), TermId.from_curie("HP:02")),
        (TermId.from_curie("HP:021"), TermId.from_curie("HP:02")),
        (TermId.from_curie("HP:022"), TermId.from_curie("HP:02")),
        (TermId.from_curie("HP:03"), TermId.from_curie("HP:1")),
    ]
    return nodes, edges


@pytest.fixture
def ontology_graph() -> OntologyGraph:
    factory = IncrementalCsrGraphFactory()
    _, edges = make_nodes_and_edges()
    return factory.create_graph(edges)


class TestHierarchicalSimilaritySorting:
    @pytest.fixture
    def ic_sim_sorting(self, ontology_graph: OntologyGraph) -> HierarchicalIcTermIdSorting:
        ics = {
            "HP:1": 0.0,
            "HP:01": 1.0,
            "HP:010": 2.0,
            "HP:011": 2.3,
            "HP:0110": 3.2,
            "HP:02": 1.5,
            "HP:020": 2.4,
            "HP:021": 2.5,
            "HP:022": 2.8,
            "HP:03": 2.0,
        }

        def ic_source(term_id: TermId) -> float:
            return ics.get(term_id.value, 0.0)

        return HierarchicalIcTermIdSorting(ontology_graph, ic_source)

    @pytest.mark.parametrize(
        "curies, expected",
        [
            (
                (
                    "HP:021",
                    "HP:03",
                    "HP:01",
                    "HP:02",
                    "HP:1",
                    "HP:022",
                    "HP:0110",
                    "HP:020",
                ),
                (
                    "HP:0110",
                    "HP:01",
                    "HP:020",
                    "HP:022",
                    "HP:02",
                    "HP:021",
                    "HP:1",
                    "HP:03",
                ),
            ),
            # Does not preserve order at the moment :'(.
            (
                (
                    "HP:0110",
                    "HP:01",
                    "HP:020",
                    "HP:022",
                    "HP:02",
                    "HP:021",
                    "HP:1",
                    "HP:03",
                ),
                (
                    "HP:01",
                    "HP:0110",
                    "HP:021",
                    "HP:02",
                    "HP:022",
                    "HP:020",
                    "HP:03",
                    "HP:1",
                ),
            ),
            (("HP:021",), ("HP:021",)),
        ],
    )
    def test_argsort(self, curies, expected, ic_sim_sorting: HierarchicalIcTermIdSorting):
        term_ids = tuple(map(TermId.from_curie, curies))

        term_indices = ic_sim_sorting.argsort(term_ids)

        ordered = tuple(term_ids[idx].value for idx in term_indices)
        assert expected == ordered


class TestHierarchicalEdgeTermIdSorting:
    @pytest.fixture
    def edge_sim_sorting(self, ontology_graph: OntologyGraph) -> HierarchicalEdgeTermIdSorting:
        return HierarchicalEdgeTermIdSorting(ontology_graph)

    @pytest.mark.parametrize(
        "curies, expected",
        [
            (
                (
                    "HP:021",
                    "HP:03",
                    "HP:01",
                    "HP:02",
                    "HP:1",
                    "HP:022",
                    "HP:0110",
                    "HP:020",
                ),
                (
                    "HP:02",
                    "HP:021",
                    "HP:022",
                    "HP:020",
                    "HP:1",
                    "HP:03",
                    "HP:01",
                    "HP:0110",
                ),
            ),
            # Does not preserve order at the moment :'(.
            (
                (
                    "HP:0110",
                    "HP:01",
                    "HP:020",
                    "HP:022",
                    "HP:02",
                    "HP:021",
                    "HP:1",
                    "HP:03",
                ),
                (
                    "HP:1",
                    "HP:01",
                    "HP:03",
                    "HP:02",
                    "HP:020",
                    "HP:022",
                    "HP:021",
                    "HP:0110",
                ),
            ),
            (("HP:021",), ("HP:021",)),
        ],
    )
    def test_argsort(self, curies, expected, edge_sim_sorting: HierarchicalEdgeTermIdSorting):
        term_ids = tuple(map(TermId.from_curie, curies))

        term_indices = edge_sim_sorting.argsort(term_ids)

        ordered = tuple(term_ids[idx].value for idx in term_indices)
        assert ordered == expected


class TestEdgeSimilarityMeasure:
    @pytest.fixture
    def edge_sim(self, ontology_graph: OntologyGraph) -> EdgeSimilarityMeasure:
        return EdgeSimilarityMeasure(ontology_graph)

    @pytest.mark.parametrize(
        "left, right, common_ancestor, expected",
        [
            ("HP:1", "HP:1", "HP:1", 0),
            ("HP:01", "HP:01", "HP:01", 0),
            ("HP:010", "HP:0110", "HP:010", 1),
            ("HP:011", "HP:0110", "HP:011", 1),
            ("HP:01", "HP:0110", "HP:01", 2),
            ("HP:1", "HP:0110", "HP:1", 3),
            ("HP:03", "HP:01", "HP:1", 2),
            ("HP:03", "HP:02", "HP:1", 2),
            ("HP:0110", "HP:022", "HP:1", 5),
        ],
    )
    def test_compute_edge_distance(
        self,
        left: str,
        right: str,
        common_ancestor: str,
        expected: int,
        edge_sim: EdgeSimilarityMeasure,
    ):
        left = TermId.from_curie(left)
        right = TermId.from_curie(right)
        common_ancestor = TermId.from_curie(common_ancestor)

        lr_dist, lr_term_id = edge_sim.calculate_edge_distance(left, right)
        assert lr_dist == expected
        rl_dist, rl_term_id = edge_sim.calculate_edge_distance(right, left)
        assert rl_dist == expected
        assert lr_term_id == rl_term_id
        assert lr_term_id == common_ancestor



import gzip
import io
import logging
import pathlib
import ssl
import sys
import typing
import warnings
from urllib.request import urlopen

import certifi


def looks_like_url(file: str) -> bool:
    """
    Checks if the `file` looks like a URL.

    :param file: file to check.
    :return: `True` if the `file` starts with `http://` or `https://`.
    """
    return file.startswith("http://") or file.startswith("https://")


def looks_gzipped(file: str) -> bool:
    """
    Checks file suffix to determine if it looks gzipped.

    :param file: file path to check.
    :return: `True` if the `file` ends with `.gz`.
    """
    return file.endswith(".gz")


def _parse_encoding(
    encoding: typing.Optional[str],
    logger: logging.Logger,
) -> str:
    if encoding is None:
        encoding = sys.getdefaultencoding()
        logger.debug("Using default encoding '%s'", encoding)
    else:
        logger.debug("Using provided encoding '%s'", encoding)
    return encoding


def open_text_io_handle_for_reading(
    fh: typing.Union[typing.TextIO, typing.BinaryIO, pathlib.Path, str],
    timeout: int = 30,
    encoding: typing.Optional[str] = None,
) -> typing.TextIO:
    """
    Open a `io.TextIO` file handle based on `fh`.

    :param fh: a `str` or `typing.IO` to read from. If `str`, then it should be a path to a local file or a URL
      of a remote resource. Either `http` or `https` protocols are supported. The content will be uncompressed
      on the fly if the file name ends with `.gz`. If `fh` is an IO wrapper, the function ensures we get a text wrapper
      that uses given encoding.
    :param timeout: timeout in seconds used when accessing a remote resource.
    :param encoding: encoding used to decode the input or the system preferred encoding if unset.
    :return: the :class:`io.TextIO` wrapper.
    """
    logger = logging.getLogger("hpotk.util")
    encoding = _parse_encoding(encoding, logger)

    logger.debug(f"Opening {fh}")
    if isinstance(fh, (pathlib.Path, str)):
        # Can be a path to local file or URL
        fp = str(fh)
        if looks_like_url(fp):
            ctx = ssl.create_default_context(cafile=certifi.where())
            logger.debug("Looks like a URL: %s", fp)
            if not isinstance(timeout, int) or timeout <= 0:
                raise ValueError(f"If {fp} looks like URL then timeout {timeout} must be a positive `int`")
            logger.debug("Downloading with timeout=%ds", timeout)
            handle = urlopen(
                fp,
                timeout=timeout,
                context=ctx,
            )
        else:
            logger.debug("Looks like a local file: %s", fp)
            handle = open(fp, "rb")

        if looks_gzipped(fp):
            logger.debug("Looks like a gzipped data, decompressing on the fly")
            return gzip.open(handle, mode="rt", newline="", encoding=encoding)
        else:
            logger.debug("Looks like decompressed data")
            return io.TextIOWrapper(handle, encoding=encoding)
    elif isinstance(fh, typing.IO):
        if isinstance(fh, typing.BinaryIO):
            logger.debug("Looks like a binary IO")
            return io.TextIOWrapper(fh, encoding=encoding)
        elif isinstance(fh, typing.TextIO):
            return fh
        else:
            raise ValueError(f"Unexpected type {type(fh)}")
    else:
        raise ValueError(f"Unexpected type {type(fh)}")


def open_text_io_handle(
    fh: typing.Union[typing.TextIO, typing.BinaryIO, str],
    timeout: int = 30,
    encoding: typing.Optional[str] = None,
) -> typing.TextIO:
    """
    Open a `io.TextIO` file handle based on `fh`.

    :param fh: a `str` or `typing.IO` to read from. If `str`, then it should be a path to a local file or a URL
      of a remote resource. Either `http` or `https` protocols are supported. The content will be uncompressed
      on the fly if the file name ends with `.gz`. If `fh` is an IO wrapper, the function ensures we get
      a text wrapper that uses given encoding.
    :param timeout: timeout in seconds used when accessing a remote resource.
    :param encoding: encoding used to decode the input or the system preferred encoding if unset.
    :return: the :class:`io.TextIO` wrapper.
    """
    # REMOVE(v1.0.0)
    warnings.warn(
        "The method has been deprecated and will be removed in v1.0.0. Use `open_text_io_handle_for_reading` instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return open_text_io_handle_for_reading(fh, timeout, encoding)


def open_text_io_handle_for_writing(
    fh: typing.Union[typing.TextIO, typing.BinaryIO, pathlib.Path, str],
    encoding: typing.Optional[str] = None,
) -> typing.TextIO:
    """
    Open a `io.TextIO` file handle based on `fpath`.

    :param fh: a `str` with a path to a local file The content will be compressed on the fly if the file name ends
      with `.gz`.
    :param encoding: encoding used to encode the output or the system preferred encoding if unset.
    :return: a :class:`io.TextIO` wrapper.
    """
    logger = logging.getLogger("hpotk.util")
    encoding = _parse_encoding(encoding, logger)

    if isinstance(fh, (pathlib.Path, str)):
        fp = str(fh)
        if looks_gzipped(fp):
            logger.debug("Looks like gzipped data, compressing on the fly")
            return gzip.open(fh, mode="wt", newline="", encoding=encoding)
        else:
            return open(fh, "w")
    elif isinstance(fh, typing.BinaryIO):
        logger.debug("Looks like a binary IO")
        return io.TextIOWrapper(fh, encoding=encoding)
    elif isinstance(fh, typing.TextIO):
        return fh
    else:
        raise ValueError(f"Unexpected type {type(fh)}")



from ._load import load_ontology, load_minimal_ontology
from ._factory import TermFactory, MinimalTermFactory

__all__ = [
    "load_ontology",
    "load_minimal_ontology",
    "TermFactory",
    "MinimalTermFactory",
]



import unittest

import ddt
import numpy as np

from hpotk.model import TermId

from .csr import ImmutableCsrMatrix
from ._csr_graph import check_items_are_sorted
from ._csr_graph import BisectPoweredCsrOntologyGraph


@ddt.ddt
class TestFunctions(unittest.TestCase):
    @ddt.data(
        ([],),
        ([1],),
        ([1, 2],),
        (["a", "b", "c"],),
    )
    @ddt.unpack
    def test_check_items_are_sorted(self, items):
        self.assertIsNone(check_items_are_sorted(items))

    @ddt.data(
        ([2, 1], "Unsorted sequence. Item #1 (1) was less than #0 (2)"),
        ([2, 1, 3, 4, 5, 6], "Unsorted sequence. Item #1 (1) was less than #0 (2)"),
        ([1, 2, 30, 4, 5, 6], "Unsorted sequence. Item #3 (4) was less than #2 (30)"),
        ([1, 2, 3, 4, 6, 5], "Unsorted sequence. Item #5 (5) was less than #4 (6)"),
    )
    @ddt.unpack
    def test_check_items_are_sorted__unsorted_cases(self, items, message):
        with self.assertRaises(ValueError) as ctx:
            check_items_are_sorted(items)
        self.assertEqual(message, ctx.exception.args[0])


@ddt.ddt
class TestBisectPoweredCsrOntologyGraph(unittest.TestCase):
    NODES: np.ndarray
    GRAPH: BisectPoweredCsrOntologyGraph

    @classmethod
    def setUpClass(cls) -> None:
        root = TermId.from_curie("HP:1")
        curies = [
            "HP:01",
            "HP:010",
            "HP:011",
            "HP:0110",
            "HP:02",
            "HP:020",
            "HP:021",
            "HP:022",
            "HP:03",
            "HP:1",
        ]
        nodes = np.fromiter(map(TermId.from_curie, curies), dtype=object)
        row = [0, 3, 5, 7, 9, 13, 14, 15, 16, 17, 20]
        col = [1, 2, 9, 0, 3, 0, 3, 1, 2, 5, 6, 7, 9, 4, 4, 4, 9, 0, 4, 8]
        data = [-1, -1, 1, 1, -1, 1, -1, 1, 1, -1, -1, -1, 1, 1, 1, 1, 1, -1, -1, -1]
        am = ImmutableCsrMatrix(row, col, data, shape=(len(nodes), len(nodes)), dtype=int)

        cls.NODES = nodes
        cls.GRAPH = BisectPoweredCsrOntologyGraph(root, nodes, am)

    def test_root(self):
        self.assertEqual(TermId.from_curie("HP:1"), self.GRAPH.root)

    @ddt.data(
        ("HP:1", ["HP:01", "HP:02", "HP:03"]),
        ("HP:02", ["HP:020", "HP:021", "HP:022"]),
        ("HP:03", []),
    )
    @ddt.unpack
    def test_get_children(self, source, expected):
        src = TermId.from_curie(source)
        exp = set(map(TermId.from_curie, expected))

        for term_id in self.GRAPH.get_children(src):
            exp.remove(term_id)
        self.assertTrue(len(exp) == 0)

        exp = set(map(TermId.from_curie, expected))
        exp.add(src)
        for term_id in self.GRAPH.get_children(src, include_source=True):
            exp.remove(term_id)

        self.assertTrue(len(exp) == 0)

    @ddt.data(
        (
            "HP:1",
            [
                "HP:01",
                "HP:010",
                "HP:011",
                "HP:0110",
                "HP:02",
                "HP:020",
                "HP:021",
                "HP:022",
                "HP:03",
            ],
        ),
        ("HP:01", ["HP:010", "HP:011", "HP:0110"]),
        ("HP:010", ["HP:0110"]),
        ("HP:011", ["HP:0110"]),
        ("HP:0110", []),
        ("HP:02", ["HP:020", "HP:021", "HP:022"]),
        ("HP:020", []),
        ("HP:021", []),
        ("HP:022", []),
        ("HP:03", []),
    )
    @ddt.unpack
    def test_get_descendants(self, source, expected):
        src = TermId.from_curie(source)
        exp = set(map(TermId.from_curie, expected))

        for term_id in self.GRAPH.get_descendants(src):
            exp.remove(term_id)
        self.assertTrue(len(exp) == 0)

        exp = set(map(TermId.from_curie, expected))
        exp.add(src)
        for term_id in self.GRAPH.get_descendants(src, include_source=True):
            exp.remove(term_id)

        self.assertTrue(len(exp) == 0)

    @ddt.data(
        ("HP:1", []),
        ("HP:01", ["HP:1"]),
        ("HP:03", ["HP:1"]),
        ("HP:0110", ["HP:010", "HP:011"]),
    )
    @ddt.unpack
    def test_get_parents(self, source, expected):
        src = TermId.from_curie(source)
        exp = set(map(TermId.from_curie, expected))

        for term_id in self.GRAPH.get_parents(src):
            exp.remove(term_id)
        self.assertTrue(len(exp) == 0)

        exp = set(map(TermId.from_curie, expected))
        exp.add(src)
        for term_id in self.GRAPH.get_parents(src, include_source=True):
            exp.remove(term_id)

        self.assertTrue(len(exp) == 0)

    @ddt.data(
        ("HP:1", []),
        ("HP:01", ["HP:1"]),
        ("HP:0110", ["HP:010", "HP:011", "HP:01", "HP:1"]),
        ("HP:022", ["HP:02", "HP:1"]),
        ("HP:03", ["HP:1"]),
    )
    @ddt.unpack
    def test_get_ancestors(self, source, expected):
        src = TermId.from_curie(source)
        exp = set(map(TermId.from_curie, expected))

        for term_id in self.GRAPH.get_ancestors(src):
            exp.remove(term_id)
        self.assertTrue(len(exp) == 0)

        exp = set(map(TermId.from_curie, expected))
        exp.add(src)
        for term_id in self.GRAPH.get_ancestors(src, include_source=True):
            exp.remove(term_id)

        self.assertTrue(len(exp) == 0)

    @ddt.data(
        ("HP:1", False),
        ("HP:01", False),
        ("HP:010", False),
        ("HP:011", False),
        ("HP:0110", True),
        ("HP:02", False),
        ("HP:020", True),
        ("HP:021", True),
        ("HP:022", True),
        ("HP:03", True),
    )
    @ddt.unpack
    def test_is_leaf(self, source, expected):
        src = TermId.from_curie(source)
        actual = self.GRAPH.is_leaf(src)
        self.assertEqual(expected, actual)

    def test_is_leaf__unknown_source(self):
        with self.assertRaises(ValueError) as ctx:
            self.GRAPH.is_leaf(TermId.from_curie("HP:999"))
        self.assertEqual("Term ID not found in the graph: HP:999", ctx.exception.args[0])

    @ddt.data(
        ("get_parents",),
        ("get_ancestors",),
        ("get_children",),
        ("get_descendants",),
    )
    @ddt.unpack
    def test_traverse_methods_with_unknown_source(self, func_name):
        unknown = TermId.from_curie("HP:999")

        func = getattr(self.GRAPH, func_name)
        with self.assertRaises(ValueError) as ctx:
            list(func(unknown))  # We must consume the iterable!
        self.assertEqual("Term ID not found in the graph: HP:999", ctx.exception.args[0])

        func = getattr(self.GRAPH, func_name)
        with self.assertRaises(ValueError) as ctx:
            list(func(unknown, include_source=True))  # We must consume the iterable!
        self.assertEqual("Term ID not found in the graph: HP:999", ctx.exception.args[0])

    def test_contains(self):
        for node in self.NODES:
            self.assertTrue(node in self.GRAPH)

    def test_iter(self):
        expected = set(self.NODES)
        actual = set(iter(self.GRAPH))
        self.assertSetEqual(expected, actual)



import pytest

from ._impl import Aspect


class TestAspect:
    @pytest.mark.parametrize(
        "payload, expected",
        [
            (
                "P",
                Aspect.PHENOTYPE,
            ),
            (
                "H",
                Aspect.PAST_MEDICAL_HISTORY,
            ),
            (
                "I",
                Aspect.INHERITANCE,
            ),
            (
                "C",
                Aspect.ONSET_AND_CLINICAL_COURSE,
            ),
            (
                "M",
                Aspect.MODIFIER,
            ),
        ],
    )
    def test_parse(
        self,
        payload: str,
        expected: Aspect,
    ):
        actual = Aspect.parse(payload)

        assert actual == expected

    def test_parse_pony(self):
        assert Aspect.parse("Pony") is None



from hpotk.model import TermId

# Selected descendents of Mode of inheritance `HP:0000005`.

INHERITANCE_MODIFIER: TermId = TermId.from_curie("HP:0034335")
# Inheritance modifier descendents can be added if required.

#  ########################## Mendelian inheritance and its descendents #############################################  #
MENDELIAN_INHERITANCE: TermId = TermId.from_curie("HP:0034345")
"""
HP:0034345 | A mode of inheritance of diseases whose pathophysiology can be traced back to deleterious variants in a single gene.
The inheritance patterns of these single-gene (monogenic) diseases are often referred to as Mendelian in honor of Gregor Mendel. 
"""

PSEUDOAUTOSOMAL_INHERITANCE: TermId = TermId.from_curie("HP:0034339")
PSEUDOAUTOSOMAL_DOMINANT_INHERITANCE: TermId = TermId.from_curie("HP:0034340")
PSEUDOAUTOSOMAL_RECESSIVE_INHERITANCE: TermId = TermId.from_curie("HP:0034341")

X_LINKED_INHERITANCE: TermId = TermId.from_curie("HP:0001417")
X_LINKED_DOMINANT_INHERITANCE: TermId = TermId.from_curie("HP:0001423")
X_LINKED_RECESSIVE_INHERITANCE: TermId = TermId.from_curie("HP:0001419")


SEMIDOMINANT_INHERITANCE: TermId = TermId.from_curie("HP:0032113")
Y_LINKED_INHERITANCE: TermId = TermId.from_curie("HP:0001450")
AUTOSOMAL_DOMINANT_INHERITANCE: TermId = TermId.from_curie("HP:0000006")
"""
`HP:0000006` | A mode of inheritance that is observed for traits related to a gene encoded on one of the autosomes (i.e., the human chromosomes 1-22)
in which a trait manifests in heterozygotes.

In the context of medical genetics, an autosomal dominant disorder is caused when a single copy of the mutant allele is present.
Males and females are affected equally, and can both transmit the disorder with a risk of 50% for each child of inheriting the mutant allele.
"""

AUTOSOMAL_RECESSIVE_INHERITANCE: TermId = TermId.from_curie("HP:0000007")
"""
`HP:0000007` | A mode of inheritance that is observed for traits related to a gene encoded on one of the autosomes (i.e., the human chromosomes 1-22)
in which a trait manifests in individuals with two pathogenic alleles, either homozygotes (two copies of the same mutant allele)
or compound heterozygotes (whereby each copy of a gene has a distinct mutant allele). 
"""

MITOCHONDRIAL_INHERITANCE: TermId = TermId.from_curie("HP:0001427")
#  ##################################################################################################################  #

#  ########################## Multifactorial inheritance and its descendents ########################################  #
MULTIFACTORIAL_INHERITANCE: TermId = TermId.from_curie("HP:0001426")
OLIGOGENIC_INHERITANCE: TermId = TermId.from_curie("HP:0010983")
POLYGENIC_INHERITANCE: TermId = TermId.from_curie("HP:0010982")
DIGENIC_INHERITANCE: TermId = TermId.from_curie("HP:0010984")
#  ##################################################################################################################  #

SPORADIC: TermId = TermId.from_curie("HP:0003745")
SOMATIC_MOSAICISM: TermId = TermId.from_curie("HP:0001442")



import abc
import enum
import logging
import typing

logger = logging.getLogger("hpotk.ontology.io.obographs")


class NodeType(enum.Enum):
    CLASS = 1
    INDIVIDUAL = 2
    PROPERTY = 3


def create_property_value(cls, data: typing.Dict):
    if cls == PropertyValue:
        return PropertyValue(
            pred=get_attr_or_none(data, "pred"),
            val=get_attr_or_none(data, "val"),
            xrefs=data["xrefs"] if "xrefs" in data else [],
            meta=create_meta(data["meta"]) if "meta" in data else None,
        )
    elif cls == DefinitionPropertyValue:
        return DefinitionPropertyValue(
            pred=get_attr_or_none(data, "pred"),
            val=get_attr_or_none(data, "val"),
            xrefs=data["xrefs"] if "xrefs" in data else [],
            meta=create_meta(data["meta"]) if "meta" in data else None,
        )
    elif cls == BasicPropertyValue:
        return BasicPropertyValue(
            pred=get_attr_or_none(data, "pred"),
            val=get_attr_or_none(data, "val"),
            xrefs=data["xrefs"] if "xrefs" in data else [],
            meta=create_meta(data["meta"]) if "meta" in data else None,
        )
    elif cls == XrefPropertyValue:
        return XrefPropertyValue(
            lbl=get_attr_or_none(data, "lbl"),
            pred=get_attr_or_none(data, "pred"),
            val=get_attr_or_none(data, "val"),
            xrefs=data["xrefs"] if "xrefs" in data else [],
            meta=create_meta(data["meta"]) if "meta" in data else None,
        )
    elif cls == SynonymPropertyValue:
        return SynonymPropertyValue(
            synonym_type=get_attr_or_none(data, "synonymType"),
            pred=get_attr_or_none(data, "pred"),
            val=get_attr_or_none(data, "val"),
            xrefs=data["xrefs"] if "xrefs" in data else [],
            meta=create_meta(data["meta"]) if "meta" in data else None,
        )
    else:
        return None


class PropertyValue:
    def __init__(
        self,
        pred: typing.Optional[str],
        val: typing.Optional[str],
        xrefs: typing.Sequence[str],
        meta,
    ):
        self._pred = pred
        self._val = val
        self._xrefs = xrefs
        self._meta = meta

    @property
    def pred(self) -> typing.Optional[str]:
        return self._pred

    @property
    def val(self) -> typing.Optional[str]:
        return self._val

    @property
    def xrefs(self) -> typing.Sequence[str]:
        return self._xrefs

    @property
    def meta(self):
        return self._meta

    def __str__(self):
        return f"PropertyValue(pred={self.pred}, val={self.val}, xrefs={self.xrefs}, meta={self.meta})"

    def __repr__(self):
        return str(self)


class DefinitionPropertyValue(PropertyValue):
    def __str__(self):
        return f"DefinitionPropertyValue(pred={self.pred}, val={self.val}, xrefs={self.xrefs}, meta={self.meta})"

    def __repr__(self):
        return str(self)


class BasicPropertyValue(PropertyValue):
    def __str__(self):
        return f"BasicPropertyValue(pred={self.pred}, val={self.val}, xrefs={self.xrefs}, meta={self.meta})"

    def __repr__(self):
        return str(self)


class XrefPropertyValue(PropertyValue):
    def __init__(
        self,
        lbl: typing.Optional[str],
        pred: typing.Optional[str],
        val: typing.Optional[str],
        xrefs: typing.Sequence[str],
        meta,
    ):
        super().__init__(pred, val, xrefs, meta)
        self._lbl = lbl

    @property
    def lbl(self) -> typing.Optional[str]:
        return self._lbl

    def __str__(self):
        return (
            f"XrefPropertyValue(lbl={self.lbl}, pred={self.pred}, val={self.val}, xrefs={self.xrefs}, meta={self.meta})"
        )

    def __repr__(self):
        return str(self)


class SynonymPropertyValue(PropertyValue):
    def __init__(
        self,
        synonym_type: typing.Optional[str],
        pred: typing.Optional[str],
        val: typing.Optional[str],
        xrefs: typing.Sequence[str],
        meta,
    ):
        super().__init__(pred, val, xrefs, meta)
        self._synonym_type = synonym_type

    @property
    def synonym_type(self) -> typing.Optional[str]:
        return self._synonym_type

    def __str__(self):
        return (
            f"SynonymPropertyValue("
            f"synonym_type={self.synonym_type},"
            f" pred={self.pred},"
            f" val={self.val},"
            f" xrefs={self.xrefs},"
            f" meta={self.meta})"
        )

    def __repr__(self):
        return str(self)


class Meta:
    def __init__(
        self,
        definition: typing.Optional[DefinitionPropertyValue],
        synonyms: typing.Sequence[SynonymPropertyValue],
        comments: typing.Sequence[str],
        basic_property_values: typing.Sequence[BasicPropertyValue],
        xrefs: typing.Sequence[XrefPropertyValue],
        is_deprecated: bool,
    ):
        self._definition = definition
        self._synonyms = synonyms
        self._comments = comments
        self._property_values = basic_property_values
        self._xrefs = xrefs
        self._is_deprecated = is_deprecated

    @property
    def definition(self) -> typing.Optional[DefinitionPropertyValue]:
        return self._definition

    @property
    def synonyms(self) -> typing.Sequence[SynonymPropertyValue]:
        return self._synonyms

    @property
    def comments(self) -> typing.Sequence[str]:
        return self._comments

    @property
    def basic_property_values(self) -> typing.Sequence[BasicPropertyValue]:
        return self._property_values

    @property
    def xrefs(self) -> typing.Sequence[XrefPropertyValue]:
        return self._xrefs

    @property
    def is_deprecated(self) -> bool:
        return self._is_deprecated

    def __str__(self):
        return (
            f"Meta(definition={self.definition},"
            f" synonyms={self.synonyms},"
            f" comments={self.comments},"
            f" basic_property_values={self.basic_property_values},"
            f" xrefs={self.xrefs},"
            f" is_deprecated={self.is_deprecated})"
        )

    def __repr__(self):
        return str(self)


class NodeOrEdge(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def meta(self) -> typing.Optional[Meta]:
        pass


class Node(NodeOrEdge):
    def __init__(
        self,
        id: str,
        lbl: typing.Optional[str],
        node_type: NodeType,
        meta: typing.Optional[Meta],
    ):
        self._id = id
        self._lbl = lbl
        self._type = node_type
        self._meta = meta

    @property
    def id(self) -> str:
        return self._id

    @property
    def lbl(self) -> typing.Optional[str]:
        return self._lbl

    @property
    def type(self) -> NodeType:
        return self._type

    @property
    def meta(self) -> typing.Optional[Meta]:
        return self._meta

    def __str__(self):
        return f"Node(id={self.id}, lbl={self.lbl}, type={self.type} meta={self.meta})"

    def __repr__(self):
        return str(self)


class Edge(NodeOrEdge):
    def __init__(self, sub: str, pred: str, obj: str, meta: typing.Optional[Meta]):
        self._sub = sub
        self._pred = pred
        self._obj = obj
        self._meta = meta

    @property
    def sub(self) -> str:
        return self._sub

    @property
    def pred(self) -> str:
        return self._pred

    @property
    def obj(self) -> str:
        return self._obj

    @property
    def meta(self) -> typing.Optional[Meta]:
        return self._meta

    def __str__(self):
        return f"Edge(sub={self.sub}, pred={self.pred}, obj={self.obj}, meta={self.meta})"

    def __repr__(self):
        return str(self)


def get_attr_or_none(data: dict, key: str):
    return data[key] if key in data else None


def create_meta(data) -> typing.Optional[Meta]:
    definition = create_property_value(DefinitionPropertyValue, data["definition"]) if "definition" in data else None
    comments = data["comments"] if "comments" in data else []
    basic_property_values = (
        [create_property_value(BasicPropertyValue, d) for d in data["basicPropertyValues"]]
        if "basicPropertyValues" in data
        else []
    )
    synonyms = [create_property_value(SynonymPropertyValue, x) for x in data["synonyms"]] if "synonyms" in data else []
    xrefs = [create_property_value(XrefPropertyValue, x) for x in data["xrefs"]] if "xrefs" in data else []
    is_deprecated = "deprecated" in data

    return Meta(definition, synonyms, comments, basic_property_values, xrefs, is_deprecated)


def create_node(data) -> typing.Optional[Node]:
    identifier = data["id"]
    lbl = data["lbl"] if "lbl" in data else None

    if "type" not in data:
        logger.debug(f"Missing node type in {data}")
        # We cannot create a Node with no type.
        return None

    nt = data["type"]
    try:
        node_type = NodeType[nt]
    except KeyError:
        logger.debug(f"Unknown node type {nt}. Supported: {list(NodeType)}")
        return None

    if "meta" in data:
        meta = create_meta(data["meta"])
    else:
        meta = None

    return Node(identifier, lbl, node_type, meta)


def create_edge(data) -> Edge:
    if not all([key in data for key in ["sub", "pred", "obj"]]):
        raise ValueError(f"Missing required property for an edge {data}")
    meta = create_meta(data["meta"]) if "meta" in data else None
    return Edge(data["sub"], data["pred"], data["obj"], meta)



import logging
import re
import enum
import typing
from collections import defaultdict, namedtuple

from hpotk.annotations import HpoDiseases, EvidenceCode, AnnotationReference, Sex
from hpotk.model import TermId
from hpotk.ontology import MinimalOntology
from hpotk.util import open_text_io_handle_for_reading
from hpotk.constants.hpo.frequency import parse_hpo_frequency
from hpotk.constants.hpo.onset import ALL_ONSETS, ONSET
from hpotk.annotations.load._api import HpoDiseaseLoader

from ..._simple import (
    SimpleHpoDiseaseAnnotation,
    SimpleHpoDisease,
    SimpleHpoDiseases,
)

HpoAnnotationLine = namedtuple(
    "HpoAnnotationLine",
    field_names=[
        "disease_id",
        "disease_name",
        "is_negated",
        "phenotype_term_id",
        "annotation_references",
        "onset",
        "frequency",
        "sex",
        "modifiers",
        "aspect",
        "curators",
    ],
)

HPOA_VERSION_PATTERN = re.compile(r"^#(date|version): (?P<version>[\w-]+)\w?$")
HPO_PATTERN = re.compile(r"^HP:\d{7}$")
RATIO_PATTERN = re.compile(r"^(?P<numerator>\d+)/(?P<denominator>\d+)$")
PERCENTAGE_PATTERN = re.compile(r"^(?P<value>\d+\.?(\d+)?)%$")


class Ratio:
    def __init__(self):
        self.numerator = 0
        self.denominator = 0

    def merge(
        self,
        numerator: int,
        denominator: int,
    ):
        self.numerator += numerator
        self.denominator += denominator


class SimpleHpoaDiseaseLoader(HpoDiseaseLoader):
    """
    Loads HPO annotation file into :class:`HpoDiseases`.

    Any issues are reported via logger.
    """

    def __init__(
        self,
        hpo: MinimalOntology,
        cohort_size: int = 50,
        salvage_negated_frequencies: bool = False,
    ):
        if not isinstance(hpo, MinimalOntology):
            raise ValueError(f"hpo must be an instance of `MinimalOntology` but was {type(hpo)}")
        self._hpo = hpo
        self._logger = logging.getLogger(__name__)
        self._cohort_size = cohort_size
        self._salvage_negated_frequencies = salvage_negated_frequencies

    def load(
        self,
        file: typing.Union[typing.IO, str],
    ) -> HpoDiseases:
        data: typing.Mapping[str, typing.List[HpoAnnotationLine]] = defaultdict(list)
        version = None
        expecting_to_see_header_line = True
        with open_text_io_handle_for_reading(file) as fh:
            for line in fh:
                if expecting_to_see_header_line:
                    if line.startswith("#"):
                        # header
                        if line.startswith("#DatabaseID"):
                            # The older HPOA format
                            expecting_to_see_header_line = False
                        else:
                            version_matcher = HPOA_VERSION_PATTERN.match(line)
                            if version_matcher:
                                version = version_matcher.group("version")
                    else:
                        if line.startswith("database_id"):
                            expecting_to_see_header_line = False
                    continue
                else:
                    # corpus
                    hpoa, err = _parse_hpoa_line(line)
                    if hpoa is None:
                        self._logger.warning("%s in line `%s`", err, line)
                    else:
                        data[hpoa.disease_id].append(hpoa)

        diseases = []
        for disease_id, hpoa_lines in data.items():
            disease = self._assemble_hpo_disease(disease_id, hpoa_lines)
            diseases.append(disease)

        return SimpleHpoDiseases(diseases, version)

    @property
    def cohort_size(self) -> int:
        return self._cohort_size

    def _assemble_hpo_disease(
        self,
        disease_curie: str,
        hpoa_lines: typing.Sequence[HpoAnnotationLine],
    ):
        # If the hpoa_lines is empty, then there is something wrong with the `defaultdict` and the logic above.
        disease_id = TermId.from_curie(disease_curie)
        disease_name = hpoa_lines[0].disease_name
        annotations, moi, onsets = self._parse_hpo_annotations(hpoa_lines)
        return SimpleHpoDisease(disease_id, disease_name, annotations, moi, onsets)

    def _parse_hpo_annotations(
        self,
        hpoa_lines: typing.Iterable[HpoAnnotationLine],
    ) -> typing.Tuple[
        typing.Sequence[SimpleHpoDiseaseAnnotation],
        typing.Collection[TermId],
        typing.Collection[TermId],
    ]:
        line_by_phenotype: typing.Mapping[str, typing.List[HpoAnnotationLine]] = defaultdict(list)

        moi = set()
        onsets = set()
        for hpoa in hpoa_lines:
            if hpoa.aspect == Aspect.PHENOTYPE:
                # Several HPOA lines may correspond to a single phenotype feature
                line_by_phenotype[hpoa.phenotype_term_id].append(hpoa)
            elif hpoa.aspect == Aspect.INHERITANCE:
                moi.add(TermId.from_curie(hpoa.phenotype_term_id))
            elif hpoa.aspect == Aspect.ONSET_AND_CLINICAL_COURSE:
                term_id = TermId.from_curie(hpoa.phenotype_term_id)
                if term_id in ALL_ONSETS:
                    onsets.add(term_id)
            else:
                # TODO - handle the remaining aspect lines
                pass

        annotations = []
        for phenotype_curie, lines in line_by_phenotype.items():
            assert len(lines) != 0, "We must have at least one HPOA line for a CURIE"

            phenotype_id = TermId.from_curie(phenotype_curie)
            total_numerator, total_denominator = 0, 0
            feature_onsets = defaultdict(Ratio)
            annotation_references = set()
            modifiers = set()
            for line in lines:
                numerator, denominator = self._parse_frequency(line.is_negated, line.frequency)
                total_numerator += numerator
                total_denominator += denominator

                if len(line.onset) != 0:
                    onset = self._parse_onset(line.onset)
                    if onset is None:
                        self._logger.warning(
                            "Unable to interpret onset %s in line %s",
                            line.onset,
                            "\t".join(
                                (
                                    line.disease_id,
                                    line.disease_name,
                                    line.phenotype_term_id,
                                )
                            ),
                        )
                    else:
                        feature_onsets[onset].merge(numerator, denominator)
                        for anc in self._hpo.graph.get_ancestors(onset):
                            if anc == ONSET:
                                break
                            feature_onsets[anc].merge(numerator, denominator)

                annotation_references.update(line.annotation_references)
                modifiers.update(line.modifiers)

            ann = SimpleHpoDiseaseAnnotation(
                phenotype_id,
                numerator=total_numerator,
                denominator=total_denominator,
                onsets=((onset, (ratio.numerator, ratio.denominator)) for onset, ratio in feature_onsets.items()),
                references=annotation_references,
                modifiers=modifiers,
            )
            annotations.append(ann)

        return annotations, moi, onsets

    def _parse_frequency(
        self,
        is_negated: bool,
        frequency: str,
    ) -> typing.Tuple[int, int]:
        # An empty string is assumed to represent a case study
        if not frequency:
            numerator = 0 if is_negated else 1
            denominator = 1
            return numerator, denominator

        # HPO term, e.g. HP:0040280 (Obligate)
        hpo_match = HPO_PATTERN.match(frequency)
        if hpo_match:
            hpo_frequency = parse_hpo_frequency(frequency)
            numerator = 0 if is_negated else round(hpo_frequency.frequency * self._cohort_size)
            denominator = self._cohort_size
            return numerator, denominator

        # Ratio, e.g. 1/2
        ratio_match = RATIO_PATTERN.match(frequency)
        if ratio_match:
            denominator = int(ratio_match.group("denominator"))
            i = int(ratio_match.group("numerator"))
            if is_negated:
                if denominator == 0:
                    # fix denominator in cases like 0/0
                    denominator = self._cohort_size
                if i == 0 and self._salvage_negated_frequencies:
                    numerator = 0
                else:
                    numerator = denominator - i
            else:
                numerator = i

            return numerator, denominator

        # Percentage, e.g. 20%
        percentage_match = PERCENTAGE_PATTERN.match(frequency)
        if percentage_match:
            percentage = float(percentage_match.group("value"))
            numerator = round(percentage * self._cohort_size / 100)
            denominator = self._cohort_size
            return numerator, denominator

        raise ValueError(f"Unable to parse frequency {frequency}")

    def _parse_onset(
        self,
        onset: str,
    ) -> typing.Optional[TermId]:
        match = HPO_PATTERN.match(onset)
        if match:
            term_id = TermId.from_curie(onset)
            if self._hpo.graph.is_ancestor_of(ONSET, term_id):
                return term_id

        return None


def _parse_hpoa_line(
    line: str,
) -> typing.Tuple[
    typing.Optional[HpoAnnotationLine],
    typing.Optional[str],
]:
    fields = line.strip().split("\t")

    if len(fields) < 12:
        return None, f"Found less than 12 fields ({len(fields)})"

    disease_id = fields[0]
    disease_name = fields[1]
    is_negated = fields[2].upper() == "NOT"
    phenotype_id = fields[3]
    evidence_code = EvidenceCode.parse(fields[5])
    if evidence_code is None:
        return None, f"Invalid evidence code `{fields[5]}`"
    annotation_references = [
        AnnotationReference(TermId.from_curie(term_id), evidence_code)
        for term_id in filter(lambda t: t and not t.isspace(), fields[4].split(";"))
    ]
    onset = fields[6]

    frequency = fields[7]
    sex = Sex.parse(fields[8])

    modifiers = [
        TermId.from_curie(term_id) for term_id in filter(lambda t: t and not t.isspace(), fields[9].split(";"))
    ]
    aspect = Aspect.parse(fields[10])
    curators = [curator.strip() for curator in fields[11].split(";")]

    return (
        HpoAnnotationLine(
            disease_id,
            disease_name,
            is_negated,
            phenotype_id,
            annotation_references,
            onset,
            frequency,
            sex,
            modifiers,
            aspect,
            curators,
        ),
        None,
    )


class Aspect(enum.Enum):
    """
    An enum for the aspect column of the HPO annotation lines.
    """

    PHENOTYPE = 0
    """
    Phenotype.
    """
    INHERITANCE = 1
    """
    Inheritance.
    """
    ONSET_AND_CLINICAL_COURSE = 2
    """
    Onset and clinical course.
    """
    MODIFIER = 3
    """
    Modifier.
    """
    PAST_MEDICAL_HISTORY = 4
    """
    Past medical history.
    """

    @staticmethod
    def parse(
        value: str,
    ) -> typing.Optional["Aspect"]:
        """
        Parse :class:`Aspect` from `str` value.

        :param value: a `str` with the aspect code.
        :return: the parsed enum member or `None` if `value` is not valid :class:`Aspect` value.
        """
        value = value.upper()
        if value == "P":
            return Aspect.PHENOTYPE
        elif value == "C":
            return Aspect.ONSET_AND_CLINICAL_COURSE
        elif value == "I":
            return Aspect.INHERITANCE
        elif value == "M":
            return Aspect.MODIFIER
        elif value == "H":
            return Aspect.PAST_MEDICAL_HISTORY
        else:
            return None



import typing

import numpy as np

from hpotk.model import TermId

from ._api import OntologyGraph

from ._csr_graph import BisectPoweredCsrOntologyGraph
from .csr import ImmutableCsrMatrix


def get_toy_graph() -> typing.Tuple[typing.Sequence[TermId], OntologyGraph]:
    root = TermId.from_curie("HP:1")
    curies = [
        "HP:01",
        "HP:010",
        "HP:011",
        "HP:0110",
        "HP:02",
        "HP:020",
        "HP:021",
        "HP:022",
        "HP:03",
        "HP:1",
    ]
    nodes = np.fromiter(map(TermId.from_curie, curies), dtype=object)
    row = [0, 3, 5, 7, 9, 13, 14, 15, 16, 17, 20]
    col = [1, 2, 9, 0, 3, 0, 3, 1, 2, 5, 6, 7, 9, 4, 4, 4, 9, 0, 4, 8]
    data = [-1, -1, 1, 1, -1, 1, -1, 1, 1, -1, -1, -1, 1, 1, 1, 1, 1, -1, -1, -1]
    am = ImmutableCsrMatrix(row, col, data, shape=(len(nodes), len(nodes)), dtype=int)

    return nodes, BisectPoweredCsrOntologyGraph(root, nodes, am)



import abc
import enum
import io
import logging
import os
import shutil
import typing

from hpotk.ontology import MinimalOntology, Ontology
from hpotk.ontology.load.obographs import load_minimal_ontology, load_ontology
from hpotk.util import validate_instance


class OntologyType(enum.Enum):
    """
    Enum with the ontologies supported by the :class:`OntologyStore`.
    """

    HPO = "HPO", "HP"
    """
    Human Phenotype Ontology.
    """

    MAxO = "MAxO", "MAXO"
    """
    Medical Action Ontology.
    """

    MONDO = "MONDO", "MONDO"
    """
    Mondo Disease Ontology.
    """

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __init__(self, _: str, identifier: str):
        self._id_ = identifier

    @property
    def identifier(self) -> str:
        """
        Get a `str` with the ontology identifier (e.g. ``HP`` for HPO).

        >>> from hpotk.store import OntologyType
        >>> OntologyType.HPO.identifier
        'HP'
        >>> OntologyType.MAxO.identifier
        'MAXO'
        """
        return self._id_


class RemoteOntologyService(metaclass=abc.ABCMeta):
    """
    `RemoteOntologyService` knows how to open a :class:`typing.BinaryIO`
    for reading content of an `ontology_type` of a particular `release`.
    """

    @abc.abstractmethod
    def fetch_ontology(
        self,
        ontology_type: OntologyType,
        release: str,
    ) -> io.BufferedIOBase:
        """
        Open a connection for reading bytes of the `ontology_type` from a remote resource.

        :param ontology_type: the desired ontology kind, e.g. :class:`OntologyType.HPO`.
        :param release: a `str` with the desired ontology release.
        :return: a binary IO for reading the ontology data.
        """
        pass


class OntologyReleaseService(metaclass=abc.ABCMeta):
    """
    `OntologyReleaseService` knows how to fetch ontology release tags, such as `v2023-10-09` for HPO.
    """

    @abc.abstractmethod
    def fetch_tags(
        self,
        ontology_type: OntologyType,
    ) -> typing.Iterable[str]:
        """
        Fetch sequence of tags for an ontology.

        :param ontology_type: the target ontology type.
        :return:
        """
        pass


class OntologyStore:
    """
    `OntologyStore` stores versions of the supported ontologies.
    """

    def __init__(
        self,
        store_dir: str,
        ontology_release_service: OntologyReleaseService,
        remote_ontology_service: RemoteOntologyService,
    ):
        self._logger = logging.getLogger(__name__)
        self._store_dir = store_dir
        self._ontology_release_service = validate_instance(
            ontology_release_service,
            OntologyReleaseService,
            "ontology_release_service",
        )
        self._remote_ontology_service = validate_instance(
            remote_ontology_service,
            RemoteOntologyService,
            "remote_ontology_service",
        )

    def load_minimal_ontology(
        self,
        ontology_type: OntologyType,
        release: typing.Optional[str] = None,
        **kwargs,
    ) -> MinimalOntology:
        """
        Load a `release` of a given `ontology_type` as a minimal ontology.

        :param ontology_type: the desired ontology type, see :class:`OntologyType` for a list of supported ontologies.
        :param release: a `str` with the ontology release tag or `None` if the latest ontology should be fetched.
        :param kwargs: key-value arguments passed to the low-level loader function (currently :func:`load_minimal_ontology`).
        :return: a minimal ontology.
        """
        return self._impl_load_ontology(
            load_minimal_ontology,
            ontology_type,
            release,
            **kwargs,
        )

    def load_ontology(
        self,
        ontology_type: OntologyType,
        release: typing.Optional[str] = None,
        **kwargs,
    ) -> Ontology:
        """
        Load a `release` of a given `ontology_type` as an ontology.

        :param ontology_type: the desired ontology type, see :class:`OntologyType` for a list of supported ontologies.
        :param release: a `str` with the ontology release tag or `None` if the latest ontology should be fetched.
        :param kwargs: key-value arguments passed to the low-level loader function (currently :func:`load_ontology`).
        :return: an ontology.
        :raises ValueError: if the `release` corresponds to a non-existing ontology release.
        """
        return self._impl_load_ontology(
            load_ontology,
            ontology_type,
            release,
            **kwargs,
        )

    @property
    def store_dir(self) -> str:
        """
        Get a `str` with a platform specific absolute path to the data directory.

        The data directory points to `$HOME/.hpo-toolkit` on UNIX and `$HOME/hpo-toolkit` on Windows.
        The folder is created if it does not exist.
        """
        return self._store_dir

    def load_minimal_hpo(
        self,
        release: typing.Optional[str] = None,
    ) -> MinimalOntology:
        """
        A convenience method for loading a specific HPO release.

        :param release: an optional `str` with the desired HPO release (if `None`, the latest HPO will be provided).
        :return: a :class:`hpotk.MinimalOntology` with the HPO data.
        :raises ValueError: if the `release` corresponds to a non-existing HPO release.
        """
        return self.load_minimal_ontology(
            OntologyType.HPO,
            release=release,
            prefixes_of_interest={"HP"},
        )

    def load_hpo(
        self,
        release: typing.Optional[str] = None,
    ) -> Ontology:
        """
        A convenience method for loading a specific HPO release.

        :param release: an optional `str` with the desired HPO release (if `None`, the latest HPO will be provided).
        :return: a :class:`hpotk.Ontology` with the HPO data.
        :raises ValueError: if the `release` corresponds to a non-existing HPO release.
        """
        return self.load_ontology(
            OntologyType.HPO,
            release=release,
            prefixes_of_interest={"HP"},
        )

    def clear(
        self,
        ontology_type: typing.Optional[OntologyType] = None,
    ):
        """
        Clear all ontology resources or resources of selected `ontology_type`.

        :param ontology_type: the ontology to be cleared or `None` if resources of *all* ontologies should be cleared.
        """
        to_delete = []
        if ontology_type is None:
            to_delete.extend(os.listdir(self._store_dir))
        else:
            to_delete.append(os.path.join(self._store_dir, ontology_type.identifier))

        for item in to_delete:
            full_path = os.path.join(self._store_dir, item)
            if os.path.isdir(full_path):
                shutil.rmtree(full_path)
            else:
                os.remove(full_path)

    def resolve_store_path(
        self,
        ontology_type: OntologyType,
        release: typing.Optional[str] = None,
    ) -> str:
        """
        Resolve the path of the ontology resource (e.g. HPO `hp.json` file) within the ontology store.

        Note, the path points to the location of the ontology resource in the local filesystem.
        The path may point to a non-existing file, if the load function has not been run yet.

        **Example**

        >>> import hpotk
        >>> store = hpotk.configure_ontology_store()
        >>> store.resolve_store_path(hpotk.store.OntologyType.HPO, release='v2023-10-09')  # doctest: +SKIP
        '/home/user/.hpo-toolkit/HP/hp.v2023-10-09.json'

        :param ontology_type: the desired ontology type, see :class:`OntologyType` for a list of supported ontologies.
        :param release: an optional `str` with the desired ontology release (if `None`, the latest ontology will be provided).
        :return: a `str` with path to the ontology resource.
        """
        fdir_ontology = os.path.join(self._store_dir, ontology_type.identifier)
        if release is None:
            # Fetch the latest release tag, assuming the lexicographic tag sort order.
            release = self._fetch_latest_release_if_missing(ontology_type)

        return os.path.join(fdir_ontology, f"{ontology_type.identifier.lower()}.{release}.json")

    def _fetch_latest_release_if_missing(
        self,
        ontology_type: OntologyType,
    ) -> str:
        """
        Retrieve the latest release tag of the given `ontology_type`.

        :param ontology_type: the ontology resource of interest
        :return: a `str` with the latest ontology tag
        :raises ValueError` if unable to retrieve the latest release tag from the ontology release service
        """

        def key(val: str) -> str:
            if val.startswith("v"):
                return val[1:]
            else:
                return val

        # Fetch the latest release tag, assuming the lexicographic tag sort order and ignoring the `v` prefix.
        latest_tag = max(
            self._ontology_release_service.fetch_tags(ontology_type),
            default=None,
            key=key,
        )
        if latest_tag is None:
            raise ValueError(f"Unable to retrieve the latest tag for {ontology_type}")
        return latest_tag

    def _impl_load_ontology(
        self,
        loader_func,
        ontology_type: OntologyType,
        release: typing.Optional[str] = None,
        **kwargs,
    ):
        if release is None:
            release = self._fetch_latest_release_if_missing(ontology_type)

        fpath_ontology = self.resolve_store_path(ontology_type, release)

        # Download ontology if missing.
        if not os.path.isfile(fpath_ontology):
            fdir_ontology = os.path.dirname(fpath_ontology)
            os.makedirs(fdir_ontology, exist_ok=True)
            with (
                self._remote_ontology_service.fetch_ontology(ontology_type, release) as response,
                open(fpath_ontology, "wb") as fh_ontology,
            ):
                fh_ontology.write(response.read())

            self._logger.debug("Stored the ontology at %s", fpath_ontology)

        # Load the ontology
        return loader_func(fpath_ontology, **kwargs)



import typing

import pytest

import hpotk

from hpotk.model import Identified, ObservableFeature, TermId
from hpotk.validate import ValidationResult, ValidationLevel
from hpotk.validate import (
    AnnotationPropagationValidator,
    PhenotypicAbnormalityValidator,
    ObsoleteTermIdsValidator,
)


class SimpleFeature(Identified, ObservableFeature):
    def __init__(self, curie: str, status: bool):
        self._id = TermId.from_curie(curie)
        self._status = status

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def is_present(self) -> bool:
        return self._status


@pytest.fixture
def example_terms() -> typing.Sequence[SimpleFeature]:
    return (
        SimpleFeature("HP:0001166", True),  # Arachnodactyly
        SimpleFeature("HP:0001250", True),  # Seizure
        SimpleFeature("HP:0032648", True),  # Tubularization of Bowman capsule
        SimpleFeature("HP:0000805", False),  # excluded Enuresis
    )


class TestAnnotationPropagationValidator:
    @pytest.fixture
    def validator(self, toy_hpo: hpotk.MinimalOntology) -> AnnotationPropagationValidator:
        return AnnotationPropagationValidator(toy_hpo)

    def test_empty_input_is_allowed(self, validator: AnnotationPropagationValidator):
        results = validator.validate([])
        assert results.is_ok()

    def test_ok_input_produces_no_errors(
        self,
        validator: AnnotationPropagationValidator,
        example_terms: typing.Sequence[SimpleFeature],
    ):
        results = validator.validate(example_terms)
        assert results.is_ok()

    def test_obsolete_ancestor_produces_error(
        self,
        validator: AnnotationPropagationValidator,
        example_terms: typing.Sequence[SimpleFeature],
    ):
        example_terms = list(example_terms)
        example_terms.append(TermId.from_curie("HP:0006010"))

        results = validator.validate(example_terms)

        assert not results.is_ok()
        assert results.results[0] == ValidationResult(
            level=ValidationLevel.ERROR,
            category="annotation_propagation",
            message="Terms should not contain both present Arachnodactyly [HP:0001166] "
            f"and its present or excluded ancestor Long fingers [HP:0100807]",
        )

    @pytest.mark.parametrize(
        "ancestor_curie, ancestor_status, base_curie, base_status",
        (
            # present Long fingers vs. present Arachnodactyly
            ("HP:0100807", True, "HP:0001166", True),
            # excluded Long fingers vs. present Arachnodactyly
            ("HP:0100807", False, "HP:0001166", True),
            # The case is commented out since it is possible to have some abnormality but excluded specific abnormality
            # ("HP:0000014", True, 'HP:0000805', False),  # present Abnormality of the bladder vs. excluded Enuresis
            # excluded Abnormality of the bladder vs. excluded Enuresis
            ("HP:0000014", False, "HP:0000805", False),
        ),
    )
    def test_ancestor_presence_produces_error(
        self,
        ancestor_curie,
        ancestor_status,
        base_curie,
        base_status,
        toy_hpo: hpotk.MinimalOntology,
        validator: AnnotationPropagationValidator,
        example_terms: typing.Sequence[SimpleFeature],
    ):
        example_terms = list(example_terms)
        example_terms.append(SimpleFeature(ancestor_curie, ancestor_status))

        results = validator.validate(example_terms)

        assert not results.is_ok()
        assert 1 == len(results.results)
        first = results.results[0]
        assert ValidationLevel.ERROR == first.level
        assert "annotation_propagation" == first.category

        state = "present" if base_status else "excluded"
        assert (
            f"Terms should not contain both {state} "
            f"{toy_hpo.get_term(base_curie).name} [{toy_hpo.get_term(base_curie).identifier.value}] "
            f"and its present or excluded ancestor "
            f"{toy_hpo.get_term(ancestor_curie).name} [{ancestor_curie}]" == first.message
        )


class TestPhenotypicAbnormalityValidator:
    @pytest.fixture
    def validator(self, toy_hpo: hpotk.MinimalOntology) -> PhenotypicAbnormalityValidator:
        return PhenotypicAbnormalityValidator(toy_hpo)

    def test_ok_input_produces_no_errors(
        self,
        validator: PhenotypicAbnormalityValidator,
        example_terms: typing.Sequence[SimpleFeature],
    ):
        results = validator.validate(example_terms)
        assert results.is_ok()

    @pytest.mark.parametrize(
        "curie",
        (
            "HP:0012823",  # Clinical modifier
            "HP:0003825",  # Variable expressivity
            "HP:0003621",  # Juvenile onset
        ),
    )
    def test_presence_of_clinical_modifier_produces_error(
        self,
        curie,
        validator: PhenotypicAbnormalityValidator,
        toy_hpo: hpotk.MinimalOntology,
        example_terms: typing.Sequence[SimpleFeature],
    ):
        example_terms = list(example_terms)
        example_terms.append(SimpleFeature(curie, True))
        results = validator.validate(example_terms)

        assert not results.is_ok()
        assert results.results[0] == ValidationResult(
            level=ValidationLevel.WARNING,
            category="phenotypic_abnormality_descendant",
            message=f"{toy_hpo.get_term(curie).name} "
            f"[{toy_hpo.get_term(curie).identifier.value}] "
            f"is not a descendant of "
            "Phenotypic abnormality [HP:0000118]",
        )


class TestObsoleteTermIdsValidator:
    @pytest.fixture
    def validator(self, toy_hpo: hpotk.MinimalOntology) -> ObsoleteTermIdsValidator:
        return ObsoleteTermIdsValidator(toy_hpo)

    def test_ok_input_produces_no_errors(
        self,
        validator: ObsoleteTermIdsValidator,
        example_terms: typing.Sequence[SimpleFeature],
    ):
        results = validator.validate(example_terms)

        assert results.is_ok()

    def test_presence_of_an_obsolete_term_produces_error(
        self,
        validator: ObsoleteTermIdsValidator,
        toy_hpo: hpotk.MinimalOntology,
        example_terms: typing.Sequence[SimpleFeature],
    ):
        curie = "HP:0006010"  # obsolete Long fingers
        example_terms = list(example_terms)
        example_terms.append(SimpleFeature(curie, True))
        results = validator.validate(example_terms)

        assert not results.is_ok()
        assert results.results[0] == ValidationResult(
            level=ValidationLevel.WARNING,
            category="obsolete_term_id_is_used",
            message=f"Using the obsolete {curie} instead of "
            f"{toy_hpo.get_term(curie).identifier.value} "
            f"for {toy_hpo.get_term(curie).name}",
        )



import typing
import warnings

from hpotk.model import TermId, CURIE_OR_TERM_ID
from hpotk.util import extract_term_id
from ._api import ANNOTATED_ITEM
from ._base import AnnotationReference
from ._base import HpoDisease, HpoDiseaseAnnotation, HpoDiseases


class SimpleHpoDiseaseAnnotation(HpoDiseaseAnnotation):
    def __init__(
        self,
        identifier: TermId,
        numerator: int,
        denominator: int,
        onsets: typing.Iterable[typing.Tuple[TermId, typing.Tuple[int, int]]],
        references: typing.Iterable[AnnotationReference],
        modifiers: typing.Iterable[TermId],
    ):
        self._id = identifier
        self.check_numerator_and_denominator(numerator, denominator)
        self._numerator = numerator
        self._denominator = denominator
        self._onsets = dict(onsets)
        self._refs = tuple(references)
        self._modifiers = tuple(modifiers)

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def numerator(self) -> int:
        return self._numerator

    @property
    def denominator(self) -> int:
        return self._denominator

    @property
    def onsets(self) -> typing.Collection[TermId]:
        return self._onsets.keys()

    def onset_counts(
        self,
        onset: CURIE_OR_TERM_ID,
    ) -> typing.Optional[typing.Tuple[int, int]]:
        return self._onsets.get(extract_term_id(onset))

    @property
    def references(self) -> typing.Sequence[AnnotationReference]:
        return self._refs

    @property
    def modifiers(self) -> typing.Sequence[TermId]:
        return self._modifiers

    def __repr__(self):
        return (
            f"SimpleHpoDiseaseAnnotation("
            f"identifier={self.identifier}, "
            f"numerator={self.numerator}, "
            f"denominator={self.denominator}, "
            f"onsets={self._onsets}, "
            f"references={self.references}, "
            f"modifiers={self.modifiers})"
        )


class SimpleHpoDisease(HpoDisease):
    def __init__(
        self,
        identifier: TermId,
        name: str,
        annotations: typing.Collection[HpoDiseaseAnnotation],
        modes_of_inheritance: typing.Collection[TermId],
        onsets: typing.Collection[TermId],
    ):
        self._id = identifier
        self._name = name
        self._annotations = annotations
        self._modes_of_inheritance = modes_of_inheritance
        self._onsets = onsets

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def annotations(self) -> typing.Collection[HpoDiseaseAnnotation]:
        return self._annotations

    @property
    def modes_of_inheritance(self) -> typing.Collection[TermId]:
        return self._modes_of_inheritance

    @property
    def onsets(self) -> typing.Collection[TermId]:
        return self._onsets


class SimpleHpoDiseases(HpoDiseases):
    def __init__(
        self,
        diseases: typing.Iterable[HpoDisease],
        version: typing.Optional[str] = None,
    ):
        self._diseases = {d.identifier: d for d in diseases}
        self._version = version

    @property
    def items(self) -> typing.Collection[ANNOTATED_ITEM]:
        return self._diseases.values()

    @property
    def diseases(self) -> typing.Collection[HpoDisease]:
        # REMOVE(v1.0.0)
        warnings.warn(
            "The `diseases` property has been deprecated and will be removed in `v1.0.0`. Use `items()` instead",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self._diseases.values()

    @property
    def disease_ids(self):
        # REMOVE(v1.0.0)
        warnings.warn(
            "`disease_ids` property has been deprecated and will be removed in v1.0.0. "
            "Iterate over `item_ids()` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return list(self.item_ids())

    def __getitem__(self, item: CURIE_OR_TERM_ID) -> typing.Optional[HpoDisease]:
        if isinstance(item, TermId):
            pass
        elif isinstance(item, str):
            item = TermId.from_curie(item)
        else:
            raise ValueError(f"Expected a `str` or `TermId` but got {type(item)}")
        try:
            return self._diseases[item]
        except KeyError:
            return None

    def __iter__(self) -> typing.Iterator[HpoDiseaseAnnotation]:
        return iter(self.items)

    def __len__(self) -> int:
        return len(self._diseases)

    @property
    def version(self) -> typing.Optional[str]:
        return self._version



import io
import json
import logging
import re
import ssl
import typing
from urllib.request import urlopen

import certifi

from ._api import OntologyType, OntologyReleaseService, RemoteOntologyService


production_tag_pt = r"^v?(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})$"
"""
A tag pattern to ensure we only include the "production" tags (e.g. not `v2024-12-12X`).
"""


ONTOLOGY_CREDENTIALS = {
    OntologyType.HPO: {
        "owner": "obophenotype",
        "repo": "human-phenotype-ontology",
        "tag_pt": production_tag_pt,
    },
    OntologyType.MAxO: {
        "owner": "monarch-initiative",
        "repo": "MAxO",
        "tag_pt": production_tag_pt,
    },
    OntologyType.MONDO: {
        "owner": "monarch-initiative",
        "repo": "mondo",
        "tag_pt": production_tag_pt,
    },
}
"""
The default ontology credentials that only include HPO, MAxO, and MONDO at this time.
"""


class GitHubOntologyReleaseService(OntologyReleaseService):
    """
    `GitHubOntologyReleaseService` can fetch the ontology tags from GitHub.
    """

    def __init__(
        self,
        timeout: int = 10,
        ontology_credentials: typing.Mapping[OntologyType, typing.Mapping[str, str]] = ONTOLOGY_CREDENTIALS,
    ):
        self._logger = logging.getLogger(__name__)
        self._timeout = timeout
        self._tag_api_url = "https://api.github.com/repos/{owner}/{repo}/tags"
        self._ctx = ssl.create_default_context(cafile=certifi.where())
        self._ontology_credentials = ontology_credentials

    def fetch_tags(self, ontology_type: OntologyType) -> typing.Iterable[str]:
        if ontology_type not in self._ontology_credentials:
            raise ValueError(
                f"Ontology {ontology_type} not among "
                f"the known ontology credentials {set(self._ontology_credentials.keys())}"
            )
        credentials = self._ontology_credentials[ontology_type]

        return self._get_tag_names(
            owner=credentials["owner"],
            repo=credentials["repo"],
            tag_pt=credentials["tag_pt"],
        )

    def _get_tag_names(
        self,
        owner: str,
        repo: str,
        tag_pt: str,
    ) -> typing.Iterable[str]:
        tag_url = self._tag_api_url.format(owner=owner, repo=repo)
        self._logger.debug("Pulling tag from %s", tag_url)

        with urlopen(
            tag_url,
            timeout=self._timeout,
            context=self._ctx,
        ) as fh:
            tags = json.load(fh)

        if len(tags) == 0:
            raise ValueError("No tags could be fetched from GitHub tag API")
        else:
            self._logger.debug("Fetched %d tags", len(tags))

        return GitHubOntologyReleaseService._filter_tags(
            tag_pt=tag_pt,
            tags=(tag["name"] for tag in tags),
        )

    @staticmethod
    def _filter_tags(
        tag_pt: str,
        tags: typing.Iterable[str],
    ) -> typing.Iterable[str]:
        pattern = re.compile(tag_pt)
        return filter(
            lambda tag: pattern.match(tag),
            tags,
        )


class GitHubRemoteOntologyService(RemoteOntologyService):
    """
    `GitHubRemoteOntologyService` knows how to fetch ontology data from GitHub.

    The Obographs JSON files are fetched and only HPO is supported as of now.
    """

    def __init__(
        self,
        timeout: int = 10,
        ontology_credentials: typing.Mapping[OntologyType, typing.Mapping[str, str]] = ONTOLOGY_CREDENTIALS,
    ):
        self._logger = logging.getLogger(__name__)
        self._timeout = timeout
        self._ctx = ssl.create_default_context(cafile=certifi.where())
        self._release_url = "https://github.com/{owner}/{repo}/releases/download/{release}/{ontology_id}.json"
        self._ontology_credentials = ontology_credentials

    def fetch_ontology(
        self,
        ontology_type: OntologyType,
        release: str,
    ) -> io.BufferedIOBase:
        if ontology_type not in self._ontology_credentials:
            raise ValueError(f"Ontology {ontology_type} not among the known ontology credentials")
        credentials = self._ontology_credentials[ontology_type]

        self._logger.debug("Using %s as the ontology release", release)

        owner = credentials["owner"]
        repo = credentials["repo"]
        url = self._release_url.format(
            owner=owner,
            repo=repo,
            release=release,
            ontology_id=ontology_type.identifier.lower(),
        )
        self._logger.info("Downloading ontology from %s", url)

        return urlopen(
            url,
            timeout=self._timeout,
            context=self._ctx,
        )



"""
The `hpotk.ontology` package defines what an ontology is and provides methods for working with ontology data.
"""

from ._api import MinimalOntology, Ontology
from ._default import create_minimal_ontology, create_ontology

__all__ = [
    "MinimalOntology",
    "Ontology",
    "create_minimal_ontology",
    "create_ontology",
]



import pytest

from hpotk.model import TermId


class TestTermId:
    @pytest.mark.parametrize(
        "curie",
        (
            "HP:1234567",
            "HP_1234567",
        ),
    )
    def test_from_curie(self, curie):
        term_id = TermId.from_curie(curie)
        assert term_id.value == "HP:1234567"

    def test_from_curie__unusual_input(self):
        term_id = TermId.from_curie("SNOMEDCT_US:313307000")
        assert term_id.prefix == "SNOMEDCT_US"
        assert term_id.id == "313307000"

    @pytest.mark.parametrize(
        "curie, message",
        [
            ("HP1234567", "The CURIE HP1234567 has no colon `:` or underscore `_`"),
            (None, "Curie must not be None"),
        ],
    )
    def test_from_curie__failures(self, curie, message):
        with pytest.raises(ValueError) as cm:
            TermId.from_curie(curie)

        assert message == cm.value.args[0]

    @pytest.mark.parametrize(
        "curie, prefix, id",
        (["HP:1234567", "HP", "1234567"], ["HP_1234567", "HP", "1234567"]),
    )
    def test_properties(self, curie, prefix, id):
        term_id = TermId.from_curie(curie)

        assert prefix == term_id.prefix
        assert id == term_id.id
        assert f"{prefix}:{id}" == term_id.value

    @pytest.mark.parametrize(
        "left, right",
        (
            ["HP:1234567", "HP:1234567"],
            ["HP:1234567", "HP_1234567"],
            ["HP_1234567", "HP:1234567"],
            ["HP_1234567", "HP_1234567"],
        ),
    )
    def test_equality(self, left, right):
        left = TermId.from_curie(left)
        right = TermId.from_curie(right)

        assert left == right

    @pytest.mark.parametrize(
        "left, right",
        (
            ["HP:1234567", "HP:1234567"],
            ["HP:1234567", "HP_1234567"],
            ["HP_1234567", "HP:1234567"],
            ["HP_1234567", "HP_1234567"],
        ),
    )
    def test_hash(self, left: str, right: str):
        left = TermId.from_curie(left)
        right = TermId.from_curie(right)

        assert hash(left) == hash(right)

    @pytest.mark.parametrize(
        "left, right, is_gt, is_eq, is_lt",
        (
            # First compare by prefix
            ["AA:0000000", "BB:0000000", False, False, True],
            ["BB:0000000", "BB:0000000", False, True, False],
            ["CC:0000000", "BB:0000000", True, False, False],
            # Then by value
            ["HP:0000001", "HP:0000002", False, False, True],
            ["HP:0000002", "HP:0000002", False, True, False],
            ["HP:0000003", "HP:0000002", True, False, False],
        ),
    )
    def test_comparison(self, left, right, is_gt, is_eq, is_lt):
        left = TermId.from_curie(left)
        right = TermId.from_curie(right)
        assert (left > right) == is_gt
        assert (left == right) == is_eq
        assert (left < right) == is_lt



import abc
import typing

from hpotk.model import (
    ID,
    CURIE_OR_TERM_ID_OR_IDENTIFIED,
    TERM,
    MINIMAL_TERM,
    Versioned,
)
from hpotk.graph import GraphAware


class MinimalOntology(typing.Generic[ID, MINIMAL_TERM], GraphAware[ID], Versioned, metaclass=abc.ABCMeta):
    """
    `MinimalOntology` is a data structure for representing the ontology terms
    and the ontology hierarchy.

    The typical way to load the ontology is by parsing Obographs JSON file
    using :class:`hpotk.util.store.OntologyStore`, see :ref:`rstload-ontology`
    section for more info.

    Here we will load a toy HPO shipped with the documentation:

    >>> import os
    >>> import hpotk
    >>> fpath_hpo = os.path.join('docs', 'data', 'hp.toy.json')
    >>> hpo = hpotk.load_minimal_ontology(fpath_hpo)

    The ontology includes the following:

    * ontology hierarchy as :class:`hpotk.graph.OntologyGraph`
    * ontology terms as :class:`hpotk.model.MinimalTerm`
    * the metadata, such as the ontology version

    The ontology acts as a Python container of term IDs,
    we can check if a term is in the ontology as:

    >>> seizure_curie = 'HP:0001250'
    >>> seizure_curie in hpo
    True

    This works for term IDs too:



    >>> seizure_id = hpotk.TermId.from_curie(seizure_curie)
    >>> seizure_id in hpo
    True

    The ontology has length - the number of *primary* terms:

    >>> len(hpo)
    393

    .. note::

      The toy HPO has only 393 terms. Real-life HPO has much more terms.

    The terms of `MinimalOntology` are instances of :class:`hpotk.model.MinimalTerm`.
    """

    @property
    @abc.abstractmethod
    def term_ids(self) -> typing.Iterator[ID]:
        """
        Get an iterator over term IDs of the primary AND obsolete ontology terms.
        """
        pass

    @property
    @abc.abstractmethod
    def terms(self) -> typing.Iterator[MINIMAL_TERM]:
        """
        Get an iterator over current terms (*not* obsolete terms).
        """
        pass

    @abc.abstractmethod
    def get_term(self, term_id: CURIE_OR_TERM_ID_OR_IDENTIFIED) -> typing.Optional[MINIMAL_TERM]:
        """
        Get the current term for a `term_id`.

        .. testsetup::

          >>> import os, hpotk
          >>> hpo = hpotk.load_minimal_ontology(os.path.join('docs', 'data', 'hp.toy.json'))

        >>> seizure = hpo.get_term('HP:0001250')
        >>> seizure.name
        'Seizure'

        :param term_id: a CURIE `str` (e.g. 'HP:1234567'), a :class:`hpotk.model.TermId` or
         an :class:`hpotk.model.Identified` entity that represents a *current* or an *obsolete* term.
        :return: the current term or `None` if the ontology does not contain the term ID.
        """
        pass

    def get_term_name(self, term_id: CURIE_OR_TERM_ID_OR_IDENTIFIED) -> typing.Optional[str]:
        """
        Get the name of the term with a `term_id`.

        .. testsetup::

          >>> import os, hpotk
          >>> hpo = hpotk.load_minimal_ontology(os.path.join('docs', 'data', 'hp.toy.json'))

        >>> seizure_name = hpo.get_term_name('HP:0001250')
        >>> seizure_name
        'Seizure'

        :param term_id: a CURIE `str` (e.g. 'HP:1234567'), a :class:`hpotk.model.TermId` or
         an :class:`hpotk.model.Identified` entity that represents a *current* or an *obsolete* term.
        :return: name of the term if the term is in ontology or `None` otherwise.
        """
        term = self.get_term(term_id)
        return term.name if term else None

    def __contains__(self, term_id: CURIE_OR_TERM_ID_OR_IDENTIFIED) -> bool:
        """
        Test if the ontology contains a `term_id`.

        Use :func:`get_term` if you want to use the corresponding term apart from knowing that it is there.

        .. testsetup::

          >>> import os, hpotk
          >>> hpo = hpotk.load_minimal_ontology(os.path.join('docs', 'data', 'hp.toy.json'))

        >>> 'HP:0001250' in hpo  # CURIE
        True

        >>> term_id = hpotk.TermId.from_curie('HP:0001250')
        >>> term_id in hpo
        True

        >>> seizure = hpo.get_term('HP:0001250')
        >>> seizure in hpo
        True

        :param term_id: a CURIE `str` (e.g. HP:1234567'), a :class:`hpotk.model.TermId` or
         an :class:`hpotk.model.Identified` entity that represents a *current* or an *obsolete* term.
        :return: `True` if the term is in the ontology and `False` otherwise.
        """
        return self.get_term(term_id) is not None

    @abc.abstractmethod
    def __len__(self) -> int:
        """
        Get the number of the primary (non-obsolete) terms in the ontology.

        .. testsetup::

        >>> import os, hpotk
        >>> hpo = hpotk.load_minimal_ontology(os.path.join('docs', 'data', 'hp.toy.json'))

        >>> len(hpo)
        393

        :return: the number of primary terms
        """
        pass


class Ontology(MinimalOntology[ID, TERM], metaclass=abc.ABCMeta):
    """
    An ontology with all information available for terms.

    The terms `Ontology` are instances of :class:`hpotk.model.Term`.
    """

    pass



import pytest

from hpotk.model import TermId

from ._simple import SimpleHpoDiseaseAnnotation


class TestSimpleHpoAnnotation:
    @pytest.mark.parametrize(
        "numerator,denominator,msg",
        [
            (-1, 1, "Numerator -1 must be a non-negative `int`"),
            (1, 0, "Denominator 0 must be a positive `int`"),
            (1, -1, "Denominator -1 must be a positive `int`"),
        ],
    )
    def test_errors(self, numerator, denominator, msg):
        tid = TermId.from_curie("HP:1234567")
        with pytest.raises(ValueError) as eh:
            SimpleHpoDiseaseAnnotation(tid, numerator, denominator, {}, (), ())

        assert eh.value.args == (msg,)



from ._ic import calculate_ic_for_annotated_items
from ._model import AnnotationIcContainer, SimilarityContainer
from ._resnik import precalculate_ic_mica_for_hpo_concept_pairs

__all__ = [
    "calculate_ic_for_annotated_items",
    "AnnotationIcContainer",
    "SimilarityContainer",
    "precalculate_ic_mica_for_hpo_concept_pairs",
]



from hpotk.model import *


class TestMinimalTerm:
    def test_equal_terms_are_equal(self):
        one = MinimalTerm.create_minimal_term(
            TermId.from_curie("HP:1234567"),
            "First",
            [TermId.from_curie("HP:1111111")],
            False,
        )
        other = MinimalTerm.create_minimal_term(
            TermId.from_curie("HP:1234567"),
            "First",
            [TermId.from_curie("HP:1111111")],
            False,
        )
        assert one == other

    def test_unequal_terms_are_not_equal(self):
        one = MinimalTerm.create_minimal_term(
            TermId.from_curie("HP:1234567"),
            "First",
            [TermId.from_curie("HP:1111111")],
            False,
        )
        other = MinimalTerm.create_minimal_term(
            TermId.from_curie("HP:1234567"),
            "Second",
            [TermId.from_curie("HP:1111111")],
            False,
        )
        assert one != other


class TestTerm:
    def test_equal_terms_are_equal(self):
        one_id = TermId.from_curie("HP:1234567")
        one_syn = Synonym(
            "F",
            SynonymCategory.EXACT,
            None,
            [TermId.from_curie("ORCID:1234-5678-1234-5678")],
        )
        one = Term.create_term(
            identifier=one_id,
            name="First",
            alt_term_ids=[TermId.from_curie("HP:1111111")],
            is_obsolete=False,
            definition="First term definition",
            comment="First comment",
            synonyms=[one_syn],
            xrefs=[TermId.from_curie("SNOMED_CT:123456")],
        )
        two_id = TermId.from_curie("HP:1234567")
        two = Term.create_term(
            identifier=two_id,
            name="First",
            alt_term_ids=[TermId.from_curie("HP:1111111")],
            is_obsolete=False,
            definition=Definition("First term definition", ()),
            comment="First comment",
            synonyms=[one_syn],
            xrefs=[TermId.from_curie("SNOMED_CT:123456")],
        )
        assert one == two

    def test_not_equal_terms_are_not_equal(self):
        one_id = TermId.from_curie("HP:1111111")
        one = Term.create_term(
            identifier=one_id,
            name="First",
            alt_term_ids=[TermId.from_curie("HP:2222222")],
            is_obsolete=False,
            definition="First term definition",
            comment="Other comment",
            synonyms=None,
            xrefs=None,
        )
        two_id = TermId.from_curie("HP:1111111")
        two = Term.create_term(
            identifier=two_id,
            name="First",
            alt_term_ids=[TermId.from_curie("HP:2222222")],
            is_obsolete=False,
            definition="First term definition",
            comment="First comment",
            synonyms=None,
            xrefs=None,
        )
        assert one != two

    def test_current_obsolete_synonyms(self):
        current_one = Synonym("A", SynonymCategory.EXACT, SynonymType.LAYPERSON_TERM, None)
        current_two = Synonym("B", SynonymCategory.EXACT, SynonymType.LAYPERSON_TERM, None)
        obsolete_two = Synonym("C", SynonymCategory.EXACT, SynonymType.OBSOLETE_SYNONYM, None)
        term = Term.create_term(
            identifier=TermId.from_curie("HP:1111111"),
            name="First",
            alt_term_ids=[TermId.from_curie("HP:2222222")],
            is_obsolete=False,
            definition="First term definition",
            comment="Other comment",
            synonyms=[current_one, current_two, obsolete_two],
            xrefs=None,
        )

        current = list(map(lambda s: s.name, term.current_synonyms()))
        assert current == ["A", "B"]
        obsolete = list(map(lambda s: s.name, term.obsolete_synonyms()))
        assert obsolete == ["C"]



import abc
import logging
import re
import typing

from hpotk.model import (
    TermId,
    Term,
    MinimalTerm,
    Synonym,
    SynonymType,
    SynonymCategory,
    Definition,
)
from ._model import Node, Meta, SynonymPropertyValue

logger = logging.getLogger(__name__)

MINIMAL_TERM = typing.TypeVar("MINIMAL_TERM", bound=MinimalTerm)

OBO_PURL_PT = re.compile(r"^http://purl\.obolibrary\.org/obo/(?P<value>.+)$")
HP_VAL_PT = re.compile(r"^hp(.*)#(?P<value>.+)$")
ORCID_PT = re.compile(r".*orcid\.org/(?P<orcid>\d{4}-\d{4}-\d{4}-\d{4})$")


def create_alt_term_ids(node: Node) -> typing.List[TermId]:
    alt_term_ids = []
    if node.meta:
        for bpv in node.meta.basic_property_values:
            if bpv.pred is not None and bpv.val is not None and bpv.pred.endswith("#hasAlternativeId"):
                alt_term_ids.append(TermId.from_curie(bpv.val))
    return alt_term_ids


def create_synonyms(meta: Meta) -> typing.Optional[typing.List[Synonym]]:
    if len(meta.synonyms) == 0:
        return None
    else:
        return [parse_synonym(s) for s in meta.synonyms]


def parse_synonym(spv: SynonymPropertyValue) -> Synonym:
    synonym_category: typing.Optional[SynonymCategory] = parse_synonym_category(spv.pred)
    synonym_type: typing.Optional[SynonymType] = parse_synonym_type(spv.synonym_type)
    if len(spv.xrefs) != 0:
        xrefs = []
        for xref in spv.xrefs:
            parsed = parse_synonym_xref(xref)
            if parsed is not None:
                xrefs.append(parsed)
        xrefs = list(xrefs) if len(xrefs) != 0 else None  # shrink to fit
    else:
        xrefs = None

    return Synonym(
        name=spv.val,
        synonym_category=synonym_category,
        synonym_type=synonym_type,
        xrefs=xrefs,
    )


def parse_synonym_category(synonym_category: str) -> typing.Optional[SynonymCategory]:
    if synonym_category == "hasRelatedSynonym":
        return SynonymCategory.RELATED
    elif synonym_category == "hasExactSynonym":
        return SynonymCategory.EXACT
    elif synonym_category == "hasBroadSynonym":
        return SynonymCategory.BROAD
    elif synonym_category == "hasNarrowSynonym":
        return SynonymCategory.NARROW
    else:
        logger.debug(f"Unknown synonym category {synonym_category}")
        return None


def parse_synonym_type(synonym_type: str) -> typing.Optional[SynonymType]:
    if synonym_type is None or len(synonym_type) == 0:
        return None
    hp_obo_matcher = OBO_PURL_PT.match(synonym_type)
    if hp_obo_matcher:
        value = hp_obo_matcher.group("value")
        hp_matcher = HP_VAL_PT.match(value)
        if hp_matcher:
            value = hp_matcher.group("value")
            if value in ("layperson", "layperson term"):
                return SynonymType.LAYPERSON_TERM
            elif value == "abbreviation":
                return SynonymType.ABBREVIATION
            elif value == "uk_spelling":
                return SynonymType.UK_SPELLING
            elif value == "obsolete_synonym":
                return SynonymType.OBSOLETE_SYNONYM
            elif value == "plural_form":
                return SynonymType.PLURAL_FORM
        else:
            if value in ("HP_0034334", "allelic_requirement"):
                return SynonymType.ALLELIC_REQUIREMENT

    logger.debug(f"Unknown synonym type {synonym_type}")
    return None


def parse_synonym_xref(xref) -> typing.Optional[TermId]:
    orcid_matcher = ORCID_PT.match(xref)
    if orcid_matcher:
        return TermId.from_curie(f"ORCID:{orcid_matcher.group('orcid')}")
    else:
        try:
            # TODO: this can contain many things. Investigate..
            return TermId.from_curie(xref)
        except ValueError:
            logger.debug(f"Unable to create a synonym xref from {xref}")
            return None


def create_xrefs(meta: Meta) -> typing.Optional[typing.List[TermId]]:
    if len(meta.xrefs) == 0:
        return None
    else:
        # TODO: Expecting that all xrefs are CURIES may be a bit too naive. Investigate..
        return [TermId.from_curie(xref.val) for xref in meta.xrefs]


class ObographsTermFactory(typing.Generic[MINIMAL_TERM], metaclass=abc.ABCMeta):
    """
    Term factory turns `TermId` and obographs `Node` into an ontology term.
    """

    @abc.abstractmethod
    def create_term(self, term_id: TermId, node: Node) -> typing.Optional[MINIMAL_TERM]:
        """
        Create `MinimalTerm` or a more specific instance for `TermId` and `Node`

        The term may not be created at the discretion of the factory, in which case `None` is returned.
        """
        pass


class MinimalTermFactory(ObographsTermFactory[MinimalTerm]):
    def create_term(self, term_id: TermId, node: Node) -> typing.Optional[MinimalTerm]:
        is_obsolete = node.meta is not None and node.meta.is_deprecated
        alt_term_ids = create_alt_term_ids(node)
        return MinimalTerm.create_minimal_term(term_id, node.lbl, alt_term_ids, is_obsolete)


class TermFactory(ObographsTermFactory[Term]):
    def create_term(self, term_id: TermId, node: Node) -> typing.Optional[Term]:
        if node.meta:
            if node.meta.definition is not None:
                d = node.meta.definition.val
                xrefs = node.meta.definition.xrefs
                definition = Definition(d, xrefs)
            else:
                definition = None
            comment = ", ".join(node.meta.comments) if len(node.meta.comments) > 0 else None
            alt_term_ids = create_alt_term_ids(node)
            synonyms = create_synonyms(node.meta)
            xrefs = create_xrefs(node.meta)

            return Term.create_term(
                term_id,
                name=node.lbl,
                alt_term_ids=alt_term_ids,
                is_obsolete=node.meta.is_deprecated,
                definition=definition,
                comment=comment,
                synonyms=synonyms,
                xrefs=xrefs,
            )
        else:
            return Term.create_term(
                term_id,
                name=node.lbl,
                alt_term_ids=[],
                is_obsolete=False,
                definition=None,
                comment=None,
                synonyms=None,
                xrefs=None,
            )



import io
import os
import typing
from pathlib import Path

import pytest

import hpotk


class MockRemoteOntologyService(hpotk.store.RemoteOntologyService):
    def __init__(
        self,
        release: str,
        payload: bytes,
    ):
        self._release = release
        self._payload = payload

    def fetch_ontology(
        self,
        ontology_type: hpotk.OntologyType,
        release: typing.Optional[str] = None,
    ) -> io.BufferedIOBase:
        if release == self._release:
            return io.BytesIO(self._payload)
        else:
            raise ValueError(f"Unsupported release {release}")


class TestGitHubOntologyStoreOffline:
    @pytest.fixture(scope="class")
    def remote_ontology_service(
        self,
        fpath_toy_hpo: str,
    ) -> hpotk.store.RemoteOntologyService:
        with open(fpath_toy_hpo, "rb") as fh:
            return MockRemoteOntologyService(
                release="v2022-10-05",
                payload=fh.read(),
            )

    @pytest.fixture
    def ontology_store(
        self,
        tmp_path: Path,
        remote_ontology_service: hpotk.store.RemoteOntologyService,
    ) -> hpotk.OntologyStore:
        ontology_release_service = hpotk.store.GitHubOntologyReleaseService()
        return hpotk.store.OntologyStore(
            store_dir=str(tmp_path),
            ontology_release_service=ontology_release_service,
            remote_ontology_service=remote_ontology_service,
        )

    def test_load_minimal_hpo(
        self,
        ontology_store: hpotk.OntologyStore,
    ):
        # We start with a clean slate.
        assert len(os.listdir(ontology_store.store_dir)) == 0

        release = "v2022-10-05"
        hpo = ontology_store.load_minimal_hpo(release=release)

        assert isinstance(hpo, hpotk.MinimalOntology)
        assert hpo.version == release[1:]

        fpath_expected = os.path.join(ontology_store.store_dir, "HP", f"hp.{release}.json")
        assert os.path.isfile(fpath_expected)

    def test_load_minimal_hpo__invalid_release(
        self,
        ontology_store: hpotk.OntologyStore,
    ):
        release = "v3400-12-31"
        with pytest.raises(ValueError) as e:
            ontology_store.load_minimal_hpo(release=release)

        # We test that we get whatever exception was raised by the `RemoteOntologyService`.
        assert e.value.args[0] == f"Unsupported release {release}"

    @pytest.mark.parametrize(
        "ontology_type,release,expected_fname",
        [
            (hpotk.store.OntologyType.HPO, "v2024-04-19", "hp.v2024-04-19.json"),
            (hpotk.store.OntologyType.HPO, "v2024-04-26", "hp.v2024-04-26.json"),
            (hpotk.store.OntologyType.MAxO, "v2024-05-24", "maxo.v2024-05-24.json"),
        ],
    )
    def test_resolve_store_path(
        self,
        ontology_store: hpotk.OntologyStore,
        ontology_type: hpotk.store.OntologyType,
        release: str,
        expected_fname: str,
    ):
        actual = ontology_store.resolve_store_path(ontology_type=ontology_type, release=release)

        expected = os.path.join(ontology_store.store_dir, ontology_type.identifier, expected_fname)
        assert actual == expected

    def test_clear__everything(
        self,
        ontology_store: hpotk.OntologyStore,
    ):
        store_dir = Path(ontology_store.store_dir)

        stuff = os.listdir(store_dir)
        assert len(stuff) == 0, "The store directory is empty upon start"

        TestGitHubOntologyStoreOffline.initialize_store_dir(store_dir)

        stuff = os.listdir(store_dir)
        assert len(stuff) == 3, "The store directory now includes two folders and one file"

        ontology_store.clear()

        stuff = os.listdir(store_dir)
        assert len(stuff) == 0, "The store directory is empty after clearing everything"

    @pytest.mark.parametrize(
        "resource",
        [
            hpotk.OntologyType.HPO,
        ],
    )
    def test_clear__ontology_type(
        self,
        resource: hpotk.OntologyType,
        ontology_store: hpotk.OntologyStore,
    ):
        store_dir = Path(ontology_store.store_dir)
        stuff = os.listdir(store_dir)

        assert len(stuff) == 0, "The store directory is empty upon start"

        TestGitHubOntologyStoreOffline.initialize_store_dir(store_dir)

        stuff = os.listdir(store_dir)
        assert len(stuff) == 3, "The store directory now includes two folders and one file"

        ontology_store.clear(resource)

        stuff = os.listdir(store_dir)
        assert len(stuff) == 2

    @staticmethod
    def initialize_store_dir(store_dir: Path):
        # Make a few folders and files
        store_dir.joinpath("joe.txt").touch()  # a file

        hp_path = store_dir.joinpath("HP")  # a folder
        os.mkdir(hp_path)
        hp_path.joinpath("a.txt").touch()
        hp_path.joinpath("b.txt").touch()

        mondo_path = store_dir.joinpath("MONDO")  # another folder
        os.mkdir(mondo_path)
        mondo_path.joinpath("x.txt").touch()
        mondo_path.joinpath("y.txt").touch()

    @pytest.mark.skip("Just for manual debugging")
    def test_resolve_store_path__latest(
        self,
        ontology_store: hpotk.OntologyStore,
    ):
        latest = ontology_store.resolve_store_path(hpotk.store.OntologyType.HPO)
        print(latest)


@pytest.mark.online
class TestGitHubOntologyStoreOnline:
    """
    Tests of real-life situations.
    """

    @pytest.fixture
    def ontology_store(self, tmp_path: Path) -> hpotk.OntologyStore:
        return hpotk.OntologyStore(
            store_dir=str(tmp_path),
            ontology_release_service=hpotk.store.GitHubOntologyReleaseService(),
            remote_ontology_service=hpotk.store.GitHubRemoteOntologyService(),
        )

    def test_load_minimal_hpo(
        self,
        ontology_store: hpotk.OntologyStore,
    ):
        hpo = ontology_store.load_minimal_ontology(
            hpotk.store.OntologyType.HPO,
            release=None,
            prefixes_of_interest={
                "HP",
            },
        )
        assert hpo is not None

        assert isinstance(hpo, hpotk.MinimalOntology)
        root = hpo.graph.root
        assert isinstance(root, hpotk.TermId)
        assert root.value == "HP:0000001"

    def test_load_minimal_hpo_with_convenience(
        self,
        ontology_store: hpotk.OntologyStore,
    ):
        hpo = ontology_store.load_minimal_hpo()

        assert hpo is not None

    def test_load_minimal_maxo(self, ontology_store: hpotk.OntologyStore):
        """
        Test that we can load MAxO with a little bit of extra TLC.
        """
        maxo = ontology_store.load_minimal_ontology(
            hpotk.store.OntologyType.MAxO,
            release="v2024-05-24",
            prefixes_of_interest={"MAXO"},
        )
        assert maxo is not None

        assert isinstance(maxo, hpotk.MinimalOntology)
        assert maxo.version == "2024-05-24"

        assert len(maxo) == 1788
        assert maxo.graph.root.value == "MAXO:0000001"

    def test_load_minimal_mondo(self, ontology_store: hpotk.OntologyStore):
        """
        Test that we can load MONDO with a tiny bit of extra TLC.
        """
        mondo = ontology_store.load_minimal_ontology(
            hpotk.store.OntologyType.MONDO,
            release="v2024-06-04",
            prefixes_of_interest={"MONDO"},
        )

        assert isinstance(mondo, hpotk.MinimalOntology)
        assert mondo.version == "2024-06-04"

        assert len(mondo) == 24_260

        children = set(mondo.get_term_name(term_id) for term_id in mondo.graph.get_children(mondo.graph.root))
        assert children == {
            "disease",
            "disease characteristic",
            "disease susceptibility",
            "injury",
        }

        disease_id = "MONDO:0000001"  # `disease`
        disease = mondo.get_term(disease_id)
        assert disease is not None
        assert disease.name == "disease"

        second_children = set(mondo.get_term_name(term_id) for term_id in mondo.graph.get_children(disease_id))
        assert second_children == {"human disease", "non-human animal disease"}


class TestGitHubOntologyReleaseService:
    @pytest.fixture
    def ontology_release_service(self) -> hpotk.store.OntologyReleaseService:
        return hpotk.store.GitHubOntologyReleaseService()

    @pytest.mark.online
    def test_ontology_release_service(
        self,
        ontology_release_service: hpotk.store.OntologyReleaseService,
    ):
        actual = set(ontology_release_service.fetch_tags(hpotk.store.OntologyType.HPO))

        expected = {  # As of March 5th, 2025
            "v2021-06-13",
            "v2021-08-02",
            "v2021-10-10",
            "v2022-01-27",
            "v2022-02-14",
            "v2022-04-14",
            "v2022-06-11",
            "v2022-10-05",
            "v2022-12-15",
            "v2023-01-27",
            "v2023-04-05",
            "v2023-06-06",
            "v2023-06-17",
            "v2023-07-21",
            "v2023-09-01",
            "v2023-10-09",
            "v2024-01-11",
            "v2024-01-16",
            "v2024-02-08",
            "v2024-03-06",
            "v2024-04-03",
            "v2024-04-04",
            "v2024-04-19",
            "v2024-04-26",
            "v2024-06-25",
            "v2024-07-01",
            "v2024-08-13",
            "v2024-12-12",
            "v2025-01-16",
        }

        missing = expected.difference(actual)

        assert not missing

    def test_filtering_tags(self):
        """
        This test accesses private attributes.
        """
        tag_pt = hpotk.store._github.production_tag_pt
        tags = (
            "v2025-01-16",
            "2025-03-03",
            "WHATEVER",
        )
        filtered = tuple(hpotk.store.GitHubOntologyReleaseService._filter_tags(tag_pt=tag_pt, tags=tags))

        assert filtered == (
            "v2025-01-16",
            "2025-03-03",
        )



import typing

from hpotk.model import TermId, Identified


def extract_term_id(
    input: typing.Union[str, TermId, Identified],
) -> TermId:
    """
    Extract term ID from multiple `input` types.

    If the input is a `str`, then it is expected to represent
    a valid CURIE. A `ValueError` is raised otherwise.

    Extraction from the other types is infallible.

    Raises a `ValueError` if the `input` is not a `str`,
    a :class:`~hpotk.TermId`, or an :class:`~hpotk.model.Identified`.
    """
    if isinstance(input, str):
        return TermId.from_curie(input)
    elif isinstance(input, TermId):
        return input
    elif isinstance(input, Identified):
        return input.identifier
    else:
        raise ValueError(f"Unsupported input type {type(input)}")



"""
The `hpotk.annotations` module provides classes for working with HPO annotation data that is available for download
from `HPO release data <https://hpo.jax.org/app/data/annotations>`_.

The module contains data classes to model the annotation data. Most notable classes include :class:`HpoDiseases`,
a container of diseases, and :class:`HpoDisease` a representation of the disease data.

The :mod:`hpotk.annotations.load` module contains code for loading the annotations into from HPO annotations format.
"""

from ._api import AnnotatedItem, AnnotatedItemContainer, ANNOTATED_ITEM, ANNOTATION
from ._base import EvidenceCode, Sex, AnnotationReference
from ._base import HpoDiseaseAnnotation, HpoDisease, HpoDiseases

__all__ = [
    "HpoDiseases",
    "HpoDisease",
    "HpoDiseaseAnnotation",
    "AnnotatedItem",
    "AnnotatedItemContainer",
    "ANNOTATED_ITEM",
    "ANNOTATION",
    "EvidenceCode",
    "Sex",
    "AnnotationReference",
]



import hpotk


class TestTerms:
    """
    We only load the ontology once, and we test the properties of the loaded data.
    """

    def test_term_properties(self, toy_hpo: hpotk.Ontology):
        # Test properties of a Term
        term = toy_hpo.get_term("HP:0001626")

        assert term.identifier.value == "HP:0001626"
        assert term.name == "Abnormality of the cardiovascular system"
        definition = term.definition
        assert definition.definition == "Any abnormality of the cardiovascular system."
        assert definition.xrefs == ("HPO:probinson",)
        assert term.comment == "The cardiovascular system consists of the heart, vasculature, and the lymphatic system."
        assert not term.is_obsolete
        assert term.alt_term_ids == (hpotk.TermId.from_curie("HP:0003116"),)

        synonyms = term.synonyms
        assert len(synonyms) == 3

        one = synonyms[0]
        assert one.name == "Cardiovascular disease"
        assert one.category == hpotk.model.SynonymCategory.RELATED
        assert one.synonym_type == hpotk.model.SynonymType.LAYPERSON_TERM
        assert one.xrefs is None

        two = synonyms[1]
        assert two.name == "Cardiovascular abnormality"
        assert two.category == hpotk.model.SynonymCategory.EXACT
        assert two.synonym_type == hpotk.model.SynonymType.LAYPERSON_TERM
        assert two.xrefs is None

        three = synonyms[2]
        assert three.name == "Abnormality of the cardiovascular system"
        assert three.category == hpotk.model.SynonymCategory.EXACT
        assert three.synonym_type == hpotk.model.SynonymType.LAYPERSON_TERM
        assert three.xrefs is None

        assert term.xrefs == tuple(
            hpotk.TermId.from_curie(curie)
            for curie in (
                "UMLS:C0243050",
                "UMLS:C0007222",
                "MSH:D018376",
                "SNOMEDCT_US:49601007",
                "MSH:D002318",
            )
        )

    def test_synonym_properties(self, toy_hpo: hpotk.Ontology):
        term = toy_hpo.get_term("HP:0001627")

        synonym = term.synonyms[7]

        assert synonym.name == "Abnormally shaped heart"
        assert synonym.category == hpotk.model.SynonymCategory.EXACT
        assert synonym.synonym_type == hpotk.model.SynonymType.LAYPERSON_TERM
        assert synonym.xrefs == [hpotk.TermId.from_curie("ORCID:0000-0001-5208-3432")]



from ._api import OntologyGraph, IndexedOntologyGraph, GraphAware, OWL_THING
from ._csr_graph import SimpleCsrOntologyGraph  # REMOVE(v1.0.0)
from ._factory import CsrGraphFactory  # REMOVE(v1.0.0)
from ._factory import GraphFactory, IncrementalCsrGraphFactory, CsrIndexedGraphFactory

__all__ = [
    "OntologyGraph",
    "IndexedOntologyGraph",
    "GraphFactory",
    "IncrementalCsrGraphFactory",
    "CsrIndexedGraphFactory",
    "GraphAware",
]



"""
The `hpotk.store` package provides :class:`OntologyStore` - a class for local caching of ontology data.

The ontology store should be configured using :func:`hpotk.configure_ontology_store` function:

>>> import hpotk
>>> store = hpotk.configure_ontology_store()

The store can then be used to fetch an ontology with a given release, e.g. `v2023-10-09`:

>>> hpo = store.load_minimal_hpo(release='v2023-10-09')
>>> hpo.version
'2023-10-09'

or fetch the *latest* release by omitting the `release` argument:

>>> latest_hpo = store.load_minimal_hpo()  # doctest: +SKIP
>>> latest_hpo.version  # doctest: +SKIP
'2024-04-26'

.. note::

  The release `2024-04-26` is the latest release as of June 2024 when this documentation was written.
"""

from ._api import (
    OntologyType,
    OntologyStore,
    RemoteOntologyService,
    OntologyReleaseService,
)
from ._github import GitHubRemoteOntologyService, GitHubOntologyReleaseService
from ._config import configure_ontology_store

__all__ = [
    "configure_ontology_store",
    "OntologyType",
    "OntologyStore",
    "RemoteOntologyService",
    "OntologyReleaseService",
    "GitHubRemoteOntologyService",
    "GitHubOntologyReleaseService",
]



"""
HPO toolkit is a library for working with Human Phenotype Ontology and the HPO annotation data.
"""

from importlib.metadata import version

__version__ = version("hpo-toolkit")

from .graph import OntologyGraph, GraphAware
from .model import TermId, Term, MinimalTerm, Synonym, SynonymType, SynonymCategory
from .ontology import Ontology, MinimalOntology

from .ontology.load.obographs import load_minimal_ontology, load_ontology
from .store import OntologyType, OntologyStore, configure_ontology_store

__all__ = [
    "TermId",
    "MinimalTerm",
    "Term",
    "OntologyGraph",
    "GraphAware",
    "MinimalOntology",
    "Ontology",
    "OntologyStore",
    "OntologyType",
    "configure_ontology_store",
    "load_minimal_ontology",
    "load_ontology",
]



import typing
import gzip
import pathlib

import pytest

from hpotk.util import open_text_io_handle_for_writing, open_text_io_handle_for_reading


class TestOpenIoForWriting:
    @pytest.fixture(scope="class")
    def payload(self) -> typing.Sequence[str]:
        return ["abc\n", "def\n", "123\n"]

    @pytest.fixture(scope="function")
    def tmp_file(
        self,
        tmp_path: pathlib.Path,
    ) -> pathlib.Path:
        return tmp_path.joinpath("something.csv.gz")

    def test_works_with_path_and_hpotk(
        self,
        payload: typing.Sequence[str],
        tmp_file: pathlib.Path,
    ):
        with open_text_io_handle_for_writing(tmp_file) as fh:
            fh.writelines(payload)

        with open_text_io_handle_for_reading(tmp_file) as fh:
            actual = fh.readlines()

        assert actual == payload

    def test_works_with_path_and_gzip(
        self,
        payload: typing.Sequence[str],
        tmp_file: pathlib.Path,
    ):
        with open_text_io_handle_for_writing(tmp_file) as fh:
            fh.writelines(payload)

        with gzip.open(tmp_file, "rt") as fh:
            actual = fh.readlines()

        assert actual == payload



from . import sort

from ._extract import extract_term_id
from ._io import looks_like_url, looks_gzipped
from ._io import (
    open_text_io_handle,
    open_text_io_handle_for_reading,
    open_text_io_handle_for_writing,
)
from ._log import setup_logging, DEFAULT_LOG_FMT
from ._validate import validate_instance, validate_optional_instance

__all__ = [
    "looks_like_url",
    "looks_gzipped",
    "open_text_io_handle_for_reading",
    "open_text_io_handle_for_writing",
    "setup_logging",
    "DEFAULT_LOG_FMT",
    "extract_term_id",
    "validate_instance",
    "validate_optional_instance",
]



"""
The `hpotk.util.sort` package sorts term IDs in a meaningful way. See :ref:`sort-term-ids` section for more info.
"""

from ._api import TermIdSorting
from ._hierarchical import (
    HierarchicalSimilaritySorting,
    HierarchicalIcTermIdSorting,
    HierarchicalEdgeTermIdSorting,
)

__all__ = [
    "TermIdSorting",
    "HierarchicalEdgeTermIdSorting",
    "HierarchicalIcTermIdSorting",
    "HierarchicalSimilaritySorting",
]



"""
The `hpotk.model` package provides data structures for working with ontology data.
"""

import typing

from ._term_id import TermId
from ._base import (
    Identified,
    ObservableFeature,
    FrequencyAwareFeature,
    Named,
    Versioned,
    MetadataAware,
)
from ._term import MinimalTerm, Term, Synonym, SynonymType, SynonymCategory, Definition

# Types
ID = typing.TypeVar("ID", bound=TermId)
CURIE_OR_TERM_ID = typing.Union[str, ID]
CURIE_OR_TERM_ID_OR_IDENTIFIED = typing.Union[str, TermId, Identified]
MINIMAL_TERM = typing.TypeVar("MINIMAL_TERM", bound=MinimalTerm)
TERM = typing.TypeVar("TERM", bound=Term)

__all__ = [
    "TermId",
    "MinimalTerm",
    "Term",
    "Identified",
    "ObservableFeature",
    "FrequencyAwareFeature",
    "Named",
    "Versioned",
    "MetadataAware",
    "Synonym",
    "SynonymType",
    "SynonymCategory",
    "Definition",
    "ID",
    "CURIE_OR_TERM_ID",
    "CURIE_OR_TERM_ID_OR_IDENTIFIED",
    "MINIMAL_TERM",
    "TERM",
]



import pytest

import hpotk
from hpotk.constants.hpo.base import PHENOTYPIC_ABNORMALITY, MODE_OF_INHERITANCE
from hpotk.constants.hpo.frequency import parse_hpo_frequency
from hpotk.constants.hpo.frequency import (
    EXCLUDED,
    VERY_RARE,
    OCCASIONAL,
    FREQUENT,
    VERY_FREQUENT,
    OBLIGATE,
)


class TestHpoBase:
    def test_phenotypic_abnormality(self):
        assert PHENOTYPIC_ABNORMALITY.value == "HP:0000118"

    def test_mode_of_inheritance(self):
        assert MODE_OF_INHERITANCE.value == "HP:0000005"


class TestHpoFrequency:
    def test_term_ids(self):
        assert EXCLUDED.value == "HP:0040285"
        assert VERY_RARE.value == "HP:0040284"
        assert OCCASIONAL.value == "HP:0040283"
        assert FREQUENT.value == "HP:0040282"
        assert VERY_FREQUENT.value == "HP:0040281"
        assert OBLIGATE.value == "HP:0040280"

    @pytest.mark.parametrize(
        "term_id,frequency",
        [
            (EXCLUDED, 0.0),
            (VERY_RARE, 0.025),
            (OCCASIONAL, 0.16999999999),
            (FREQUENT, 0.545),
            (VERY_FREQUENT, 0.895),
            (OBLIGATE, 1.0),
        ],
    )
    def test_frequency(
        self,
        term_id: hpotk.TermId,
        frequency: float,
    ):
        hpo_frequency = parse_hpo_frequency(term_id)

        assert hpo_frequency is not None
        assert hpo_frequency.frequency == pytest.approx(frequency)



import re
import json
import typing
import logging

from hpotk.model import TermId, MinimalTerm, Term
from hpotk.graph import OntologyGraph
from hpotk.graph import GraphFactory, CsrIndexedGraphFactory, OWL_THING
from hpotk.ontology import (
    MinimalOntology,
    Ontology,
    create_ontology,
    create_minimal_ontology,
)
from hpotk.util import open_text_io_handle_for_reading

from ._model import create_node, create_edge
from ._model import Node, Edge, NodeType
from ._factory import (
    MinimalTermFactory,
    TermFactory,
    ObographsTermFactory,
    MINIMAL_TERM,
)

logger = logging.getLogger(__name__)

# TODO: verify PURL works for other ontologies than HPO
# TODO: thoroughly test the PURL pattern
# A pattern to match an obolibrary PURL. The PURL should is expected to have 3 parts: `prefix`, `id`, and `curie`
# The `curie` is `prefix` + '_' + `id`.
PURL_PATTERN = re.compile(r"http://purl\.obolibrary\.org/obo/(?P<curie>(?P<prefix>\w+)_(?P<id>\w+))")
DATE_PATTERN = re.compile(r".*/(?P<date>\d{4}-\d{2}-\d{2})/.*")


def load_minimal_ontology(
    file: typing.Union[typing.IO, str],
    term_factory: ObographsTermFactory[MinimalTerm] = MinimalTermFactory(),
    graph_factory: GraphFactory = CsrIndexedGraphFactory(),
    prefixes_of_interest: typing.Set[str] = {"HP"},
) -> MinimalOntology:
    return _load_impl(
        file,
        term_factory,
        graph_factory,
        prefixes_of_interest,
        create_minimal_ontology,
    )


def load_ontology(
    file: typing.Union[typing.IO, str],
    term_factory: ObographsTermFactory[Term] = TermFactory(),
    graph_factory: GraphFactory = CsrIndexedGraphFactory(),
    prefixes_of_interest: typing.Set[str] = {"HP"},
) -> Ontology:
    return _load_impl(
        file,
        term_factory,
        graph_factory,
        prefixes_of_interest,
        create_ontology,
    )


def _load_impl(
    file: typing.Union[typing.IO, str],
    term_factory: ObographsTermFactory[MINIMAL_TERM],
    graph_factory: GraphFactory,
    prefixes_of_interest: typing.Set[str],
    ontology_creator,
):
    obograph = get_obographs_graph(file)
    logger.debug("Extracting ontology terms")
    id_to_term_id, terms = extract_terms(
        obograph["nodes"],
        term_factory,
        prefixes_of_interest=prefixes_of_interest,
    )
    logger.debug("Creating the edge list")
    edge_list = create_edge_list(obograph["edges"], id_to_term_id)
    logger.debug("Building ontology graph")
    ontology_graph: OntologyGraph = graph_factory.create_graph(edge_list)
    if ontology_graph.root == OWL_THING:
        # TODO: - consider adding Owl thing into terms list
        pass
    version = extract_ontology_version(obograph["meta"])
    logger.debug("Assembling the ontology")
    ontology = ontology_creator(ontology_graph, terms, version)
    logger.debug("Done")
    return ontology


def get_obographs_graph(file: typing.Union[typing.IO, str]):
    with open_text_io_handle_for_reading(file) as fh:
        document = json.load(fh)
    if not isinstance(document, dict):
        raise ValueError(f"The JSON document should have been a dict but was {type(document)}")
    if "graphs" not in document:
        raise ValueError("Did not find the `graphs` attribute in the JSON document")
    graphs = document["graphs"]
    if not isinstance(graphs, typing.Sequence):
        raise ValueError("`graphs` JSON attribute is not a sequence")
    if len(graphs) < 1:
        raise ValueError("`graphs` JSON attribute is empty")
    elif len(graphs) == 1:
        # The happy path
        return graphs[0]
    else:
        raise ValueError(f"We expect exactly 1 graph but there are {len(graphs)} graphs in the JSON document")


def extract_terms(
    nodes: typing.Iterable[dict],
    term_factory: ObographsTermFactory[MINIMAL_TERM],
    prefixes_of_interest: typing.Set[str],
) -> typing.Tuple[typing.Mapping[str, TermId], typing.Sequence[MINIMAL_TERM]]:
    curie_to_term: typing.Dict[str, TermId] = {}
    terms: typing.List[MINIMAL_TERM] = []
    for data in nodes:
        # 1) map data to `Node`
        node: typing.Optional[Node] = create_node(data)

        # 2) we only work with class Nodes
        if not node or node.type != NodeType.CLASS:
            continue

        # 3) check if PURL is OK
        curie = extract_curie_from_purl(node.id)
        if not curie:
            logger.debug("Unable to extract CURIE from PURL %s", node.id)
            continue
        term_id = TermId.from_curie(curie)
        if term_id.prefix not in prefixes_of_interest:
            logger.debug("Skipping not a term of interest %s", term_id.value)
            continue

        curie_to_term[curie] = term_id

        # 4) create the `Term`
        term = term_factory.create_term(term_id, node)
        if term:
            terms.append(term)

    return curie_to_term, terms


def create_edge_list(
    edges: typing.Iterable[typing.Dict[str, str]],
    curie_to_termid: typing.Mapping[str, TermId],
) -> typing.List[typing.Tuple[TermId, TermId]]:
    edge_list: typing.List[typing.Tuple[TermId, TermId]] = []
    for data in edges:
        edge: Edge = create_edge(data)

        # We only care about `is_a` relationships.
        if edge.pred != "is_a":
            logger.debug("Skipping edge with pred %s!='is_a'", edge.pred)
            continue

        # Get source and destination.
        src_curie = extract_curie_from_purl(edge.sub)
        if src_curie is None:
            logger.warning("Unable to extract CURIE from sub PURL %s", edge.sub)
            continue
        try:
            src: TermId = curie_to_termid[src_curie]
        except KeyError:
            logger.debug(
                "Skipping edge %s %s %s because subject %s was was not found in terms",
                edge.sub,
                edge.pred,
                edge.obj,
                edge.sub,
            )
            continue

        dest_curie = extract_curie_from_purl(edge.obj)
        if dest_curie is None:
            logger.warning("Unable to extract CURIE from obj PURL %s", edge.obj)
            continue
        try:
            dest: TermId = curie_to_termid[dest_curie]
        except KeyError:
            logger.debug(
                "Skipping edge %s %s %s because object %s was was not found in terms",
                edge.sub,
                edge.pred,
                edge.obj,
                edge.obj,
            )
            continue

        edge_list.append((src, dest))

    return edge_list


def extract_curie_from_purl(purl: str) -> typing.Optional[str]:
    """
    Parse HPO PURL (e.g. `http://purl.obolibrary.org/obo/HP_0002813`) into the CURIE (e.g. `HP_0002813`).

    Returns the CURIE `str` or `None` if the PURL is mis-formatted.
    """
    matcher = PURL_PATTERN.match(purl)
    return matcher.group("curie") if matcher else None


def extract_ontology_version(meta: dict) -> typing.Optional[str]:
    if "version" in meta:
        # A line like this:
        # 'http://purl.obolibrary.org/obo/hp/releases/2022-10-05/hp.json'
        match = DATE_PATTERN.search(meta["version"])
        if match:
            return match.group("date")
        else:
            logger.debug("Could not find a date pattern in version %s", meta["version"])
            return None
    elif "basicPropertyValues" in meta:
        for bpv in meta["basicPropertyValues"]:
            if "pred" in bpv and "val" in bpv:
                if bpv["pred"].endswith("#versionInfo"):
                    # An item like this:
                    # {
                    #   "pred": "http://www.w3.org/2002/07/owl#versionInfo",
                    #   "val": "2022-10-05"
                    # }
                    return bpv["val"]

        logger.debug("Could not find basic property value with the version info")
        return None
    else:
        logger.debug("Could not determine the ontology version")
        return None



"""
A module with term IDs of the `Onset <https://hpo.jax.org/browse/term/HP:0003674>`_
sub-hierarchy of the HPO.
"""

from hpotk.model import TermId

#  Descendents of Onset `HP:0003674`. The `#`s model the hierarchy.
ONSET: TermId = TermId.from_curie("HP:0003674")

#  #
ANTENATAL_ONSET: TermId = TermId.from_curie("HP:0030674")
#  ##
EMBRYONAL_ONSET: TermId = TermId.from_curie("HP:0011460")
#  ##
FETAL_ONSET: TermId = TermId.from_curie("HP:0011461")
#  ###
LATE_FIRST_TRIMESTER_ONSET: TermId = TermId.from_curie("HP:0034199")
SECOND_TRIMESTER_ONSET: TermId = TermId.from_curie("HP:0034198")
THIRD_TRIMESTER_ONSET: TermId = TermId.from_curie("HP:0034197")

#  #
CONGENITAL_ONSET: TermId = TermId.from_curie("HP:0003577")
NEONATAL_ONSET: TermId = TermId.from_curie("HP:0003623")
PUERPURAL_ONSET: TermId = TermId.from_curie("HP:4000040")

#  #
PEDIATRIC_ONSET: TermId = TermId.from_curie("HP:0410280")
#  ##
INFANTILE_ONSET: TermId = TermId.from_curie("HP:0003593")
CHILDHOOD_ONSET: TermId = TermId.from_curie("HP:0011463")
JUVENILE_ONSET: TermId = TermId.from_curie("HP:0003621")

#  #
ADULT_ONSET: TermId = TermId.from_curie("HP:0003581")
#  ##
YOUNG_ADULT_ONSET: TermId = TermId.from_curie("HP:0011462")
#  ###
EARLY_YOUNG_ADULT_ONSET: TermId = TermId.from_curie("HP:0025708")
INTERMEDIATE_YOUNG_ADULT_ONSET: TermId = TermId.from_curie("HP:0025709")
LATE_YOUNG_ADULT_ONSET: TermId = TermId.from_curie("HP:0025710")

#  #
MIDDLE_AGE_ONSET: TermId = TermId.from_curie("HP:0003596")
PERIMENOPAUSAL_ONSET: TermId = TermId.from_curie("HP:6000314")
POSTMENOPAUSAL_ONSET: TermId = TermId.from_curie("HP:6000315")
LATE_ONSET: TermId = TermId.from_curie("HP:0003584")


ALL_ONSETS = {
    ONSET,
    ANTENATAL_ONSET,
    EMBRYONAL_ONSET,
    FETAL_ONSET,
    LATE_FIRST_TRIMESTER_ONSET,
    SECOND_TRIMESTER_ONSET,
    THIRD_TRIMESTER_ONSET,
    CONGENITAL_ONSET,
    NEONATAL_ONSET,
    PUERPURAL_ONSET,
    PEDIATRIC_ONSET,
    INFANTILE_ONSET,
    CHILDHOOD_ONSET,
    JUVENILE_ONSET,
    ADULT_ONSET,
    YOUNG_ADULT_ONSET,
    EARLY_YOUNG_ADULT_ONSET,
    INTERMEDIATE_YOUNG_ADULT_ONSET,
    LATE_YOUNG_ADULT_ONSET,
    MIDDLE_AGE_ONSET,
    PERIMENOPAUSAL_ONSET,
    POSTMENOPAUSAL_ONSET,
    LATE_ONSET,
}
"""
A set of all members of the `Onset <https://hpo.jax.org/browse/term/HP:0003674>`_
HPO sub-hierarchy, including the root (`Onset`) term id.
"""



import abc
import typing
import warnings

from hpotk.model import (
    Identified,
    FrequencyAwareFeature,
    Versioned,
    CURIE_OR_TERM_ID_OR_IDENTIFIED,
)
from hpotk.util import extract_term_id


# #####################################################################################################################
# The file describes generic schema for annotations, annotated items, and annotated item containers.
# An example application includes phenotypic features (annotations), diseases or samples (annotated items),
# and OMIM corpus (annotated item containers).
# #####################################################################################################################


class AnnotationBase(Identified, FrequencyAwareFeature):
    pass


ANNOTATION = typing.TypeVar("ANNOTATION", bound=AnnotationBase)
"""
A world item annotation with an identifier and present or excluded state.
"""


class AnnotatedItem(
    typing.Generic[ANNOTATION],
    metaclass=abc.ABCMeta,
):
    """
    An item that can be annotated with ontology terms. For instance, a disease can be annotated with phenotypic features,
    items from HPO ontology.
    """

    @property
    @abc.abstractmethod
    def annotations(self) -> typing.Collection[ANNOTATION]:
        """
        :return: a collection of :class:`ANNOTATION` objects for the annotated item.
        """
        pass

    def present_annotations(self) -> typing.Iterable[ANNOTATION]:
        """
        :return: an iterable over *present* annotations.
        """
        return filter(lambda a: a.is_present, self.annotations)

    def absent_annotations(self) -> typing.Iterable[ANNOTATION]:
        """
        :return: an iterable over *absent* annotations.
        """
        return filter(lambda a: a.is_absent, self.annotations)

    def annotation_by_id(
        self,
        query: CURIE_OR_TERM_ID_OR_IDENTIFIED,
    ) -> typing.Optional[ANNOTATION]:
        """
        Find the annotation identified by the `query`.

        Performs a linear search and finds the *first* match.

        :param query: a `str` with CURIE, an :class:`~hpotk.TermId`, or an :class:`~hpotk.model.Identified` item (an item with an identifier).
        :return: an annotation or `None` if no such annotation exists.
        """
        term_id = extract_term_id(query)
        for ann in self.annotations:
            if ann.identifier == term_id:
                return ann
        return None


ANNOTATED_ITEM = typing.TypeVar("ANNOTATED_ITEM", bound=AnnotatedItem)
"""
World item that is annotated with an :class:`ANNOTATION`.
"""


class AnnotatedItemContainer(
    typing.Generic[ANNOTATED_ITEM],
    typing.Iterable[ANNOTATED_ITEM],
    typing.Sized,
    Versioned,
    metaclass=abc.ABCMeta,
):
    """
    Container for items that can be annotated with ontology terms.

    For instance, if OMIM disease is an item type and phenotypic feature is the annotation type,
    then a corpus of OMIM diseases corresponds to a container of annotated items.
    """

    @property
    def items(self) -> typing.Collection[ANNOTATED_ITEM]:
        """
        :return: an iterable over container items.
        """
        # REMOVE(v1.0.0)
        warnings.warn(
            "`items` property has been deprecated and will be removed in v1.0.0. Iterate directly over the container.",
            DeprecationWarning,
            stacklevel=2,
        )
        return list(self)



import typing
import unittest

import ddt
import numpy as np
import pytest

import hpotk
from ._api import OntologyGraph, IndexedOntologyGraph
from ._csr_graph import BisectPoweredCsrOntologyGraph
from ._factory import CsrIndexedGraphFactory, IncrementalCsrGraphFactory
from .csr import ImmutableCsrMatrix


@pytest.fixture(scope="module")
def edges() -> typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]:
    # Edges for a graph with a single root node: HP:1
    return (
        # source -> destination
        (hpotk.TermId.from_curie("HP:01"), hpotk.TermId.from_curie("HP:1")),
        (hpotk.TermId.from_curie("HP:010"), hpotk.TermId.from_curie("HP:01")),
        (hpotk.TermId.from_curie("HP:011"), hpotk.TermId.from_curie("HP:01")),
        # This one is multi-parent.
        (hpotk.TermId.from_curie("HP:0110"), hpotk.TermId.from_curie("HP:010")),
        (hpotk.TermId.from_curie("HP:0110"), hpotk.TermId.from_curie("HP:011")),
        (hpotk.TermId.from_curie("HP:02"), hpotk.TermId.from_curie("HP:1")),
        (hpotk.TermId.from_curie("HP:020"), hpotk.TermId.from_curie("HP:02")),
        (hpotk.TermId.from_curie("HP:021"), hpotk.TermId.from_curie("HP:02")),
        (hpotk.TermId.from_curie("HP:022"), hpotk.TermId.from_curie("HP:02")),
        (hpotk.TermId.from_curie("HP:03"), hpotk.TermId.from_curie("HP:1")),
    )


@pytest.fixture(scope="module")
def indexed_graph(
    edges: typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]],
) -> IndexedOntologyGraph:
    factory = CsrIndexedGraphFactory()
    return factory.create_graph(edges)


@pytest.fixture(scope="module")
def base_graph(
    edges: typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]],
) -> OntologyGraph:
    factory = IncrementalCsrGraphFactory()
    return factory.create_graph(edges)


@pytest.mark.parametrize(
    "source, expected",
    [
        ("HP:1", {"HP:01", "HP:02", "HP:03"}),
        ("HP:01", {"HP:010", "HP:011"}),
        ("HP:010", {"HP:0110"}),
        ("HP:011", {"HP:0110"}),
        ("HP:0110", {}),
        ("HP:02", {"HP:020", "HP:021", "HP:022"}),
        ("HP:03", {}),
    ],
)
class TestGetChildren:
    def test_get_children__indexed_graph(self, indexed_graph: IndexedOntologyGraph, source, expected):
        actual = set(indexed_graph.get_children(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    def test_get_children__base_graph(self, base_graph: OntologyGraph, source, expected):
        actual = set(base_graph.get_children(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)


@pytest.mark.parametrize(
    "source, expected",
    [
        (
            "HP:1",
            {
                "HP:01",
                "HP:010",
                "HP:011",
                "HP:0110",
                "HP:02",
                "HP:020",
                "HP:021",
                "HP:022",
                "HP:03",
            },
        ),
        ("HP:01", {"HP:010", "HP:011", "HP:0110"}),
        ("HP:010", {"HP:0110"}),
        ("HP:011", {"HP:0110"}),
        ("HP:0110", set()),
        ("HP:02", {"HP:020", "HP:021", "HP:022"}),
        ("HP:020", set()),
        ("HP:021", set()),
        ("HP:022", set()),
        ("HP:03", set()),
    ],
)
class TestGetDescendants:
    def test_get_descendants__indexed_graph(self, indexed_graph: IndexedOntologyGraph, source, expected):
        actual = set(indexed_graph.get_descendants(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    def test_get_descendants__base_graph(self, base_graph: OntologyGraph, source, expected):
        actual = set(base_graph.get_descendants(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)


@pytest.mark.parametrize(
    "source, expected",
    [
        ("HP:1", {}),
        ("HP:01", {"HP:1"}),
        ("HP:010", {"HP:01"}),
        ("HP:0110", {"HP:010", "HP:011"}),
        ("HP:03", {"HP:1"}),
    ],
)
class TestGetParents:
    def test_get_parents__indexed_graph(self, indexed_graph: IndexedOntologyGraph, source, expected):
        children = set(indexed_graph.get_parents(source))

        assert children == set(hpotk.TermId.from_curie(curie) for curie in expected)

    def test_get_parents__base_graph(self, base_graph: OntologyGraph, source, expected):
        actual = set(base_graph.get_parents(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)


@pytest.mark.parametrize(
    "source, expected",
    [
        ("HP:1", set()),
        ("HP:01", {"HP:1"}),
        ("HP:0110", {"HP:010", "HP:011", "HP:01", "HP:1"}),
        ("HP:022", {"HP:02", "HP:1"}),
        ("HP:03", {"HP:1"}),
    ],
)
class TestGetAncestors:
    def test_indexed_graph(self, indexed_graph: IndexedOntologyGraph, source, expected):
        actual = set(indexed_graph.get_ancestors(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    def test_base_graph(self, base_graph: OntologyGraph, source, expected):
        actual = set(base_graph.get_ancestors(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)


@pytest.mark.parametrize(
    "sub, obj, expected",
    [
        # True examples
        ("HP:1", "HP:01", True),
        ("HP:01", "HP:010", True),
        ("HP:01", "HP:011", True),
        ("HP:010", "HP:0110", True),
        ("HP:011", "HP:0110", True),
        ("HP:1", "HP:02", True),
        ("HP:02", "HP:020", True),
        ("HP:02", "HP:021", True),
        ("HP:02", "HP:022", True),
        ("HP:1", "HP:03", True),
        # False examples
        ("HP:1", "HP:1", False),
        ("HP:02", "HP:01", False),
        ("HP:01", "HP:020", False),
        ("HP:0110", "HP:03", False),
    ],
)
class TestIsParentOf:
    def test_indexed_graph(self, indexed_graph: IndexedOntologyGraph, sub, obj, expected):
        assert indexed_graph.is_parent_of(sub, obj) is expected

    def test_base_graph(self, base_graph: OntologyGraph, sub, obj, expected):
        assert base_graph.is_parent_of(sub, obj) is expected


@pytest.mark.parametrize(
    "sub, obj, expected",
    [
        # True examples
        ("HP:1", "HP:010", True),
        ("HP:1", "HP:011", True),
        ("HP:1", "HP:0110", True),
        ("HP:01", "HP:0110", True),
        ("HP:1", "HP:020", True),
        ("HP:1", "HP:021", True),
        ("HP:1", "HP:022", True),
        # False examples
        ("HP:1", "HP:1", False),
        ("HP:01", "HP:1", False),
    ],
)
class TestIsAncestorOf:
    def test_indexed_graph(self, indexed_graph: IndexedOntologyGraph, sub, obj, expected):
        assert indexed_graph.is_ancestor_of(sub, obj) is expected

    def test_base_graph(self, base_graph: OntologyGraph, sub, obj, expected):
        assert base_graph.is_ancestor_of(sub, obj) is expected


@pytest.mark.parametrize(
    "sub, obj, expected",
    [
        # True examples
        ("HP:01", "HP:1", True),
        ("HP:010", "HP:01", True),
        ("HP:011", "HP:01", True),
        ("HP:0110", "HP:010", True),
        ("HP:0110", "HP:011", True),
        ("HP:02", "HP:1", True),
        ("HP:020", "HP:02", True),
        ("HP:021", "HP:02", True),
        ("HP:022", "HP:02", True),
        ("HP:03", "HP:1", True),
        # False examples
        ("HP:1", "HP:1", False),
        ("HP:01", "HP:02", False),
        ("HP:020", "HP:01", False),
        ("HP:03", "HP:0110", False),
    ],
)
class TestIsChildOf:
    def test_indexed_graph(self, indexed_graph: IndexedOntologyGraph, sub, obj, expected):
        assert indexed_graph.is_child_of(sub, obj) is expected

    def test_base_graph(self, base_graph: OntologyGraph, sub, obj, expected):
        assert base_graph.is_child_of(sub, obj) is expected


@pytest.mark.parametrize(
    "sub, obj, expected",
    [
        # True examples
        ("HP:010", "HP:1", True),
        ("HP:011", "HP:1", True),
        ("HP:0110", "HP:1", True),
        ("HP:0110", "HP:01", True),
        ("HP:020", "HP:1", True),
        ("HP:021", "HP:1", True),
        ("HP:022", "HP:1", True),
        # False examples
        ("HP:1", "HP:1", False),
        ("HP:1", "HP:01", False),
    ],
)
class TestIsDescendantOf:
    def test_indexed_graph(self, indexed_graph: IndexedOntologyGraph, sub, obj, expected):
        assert indexed_graph.is_descendant_of(sub, obj) is expected

    def test_base_graph(self, base_graph: OntologyGraph, sub, obj, expected):
        assert base_graph.is_descendant_of(sub, obj) is expected


@pytest.mark.parametrize(
    "query, expected",
    [
        ("HP:1", False),
        ("HP:01", False),
        ("HP:010", False),
        ("HP:011", False),
        ("HP:0110", True),
        ("HP:02", False),
        ("HP:020", True),
        ("HP:021", True),
        ("HP:022", True),
        ("HP:03", True),
    ],
)
class TestIsLeaf:
    def test_indexed_graph(self, indexed_graph: IndexedOntologyGraph, query, expected):
        assert indexed_graph.is_leaf(query) is expected

    def test_base_graph(self, base_graph: OntologyGraph, query, expected):
        assert base_graph.is_leaf(query) is expected


class TestOntologyGraphQueries:
    @pytest.mark.parametrize(
        "sub, obj, expected",
        [
            ("HP:1", "HP:1", True),
            ("HP:1", "HP:01", True),
            ("HP:1", "HP:010", False),
        ],
    )
    def test_is_parent_of_or_equal_to(
        self,
        base_graph: OntologyGraph,
        sub: str,
        obj: str,
        expected: bool,
    ):
        assert (
            base_graph.is_parent_of_or_equal_to(
                sub,
                obj,
            )
            is expected
        )

    @pytest.mark.parametrize(
        "sub, obj, expected",
        [
            ("HP:1", "HP:1", True),
            ("HP:1", "HP:01", True),
            ("HP:1", "HP:010", True),
        ],
    )
    def test_is_ancestor_of_or_equal_to(
        self,
        base_graph: OntologyGraph,
        sub: str,
        obj: str,
        expected: bool,
    ):
        assert (
            base_graph.is_ancestor_of_or_equal_to(
                sub,
                obj,
            )
            is expected
        )

    @pytest.mark.parametrize(
        "sub, obj, expected",
        [
            ("HP:1", "HP:1", True),
            ("HP:01", "HP:1", True),
            ("HP:010", "HP:1", False),
        ],
    )
    def test_is_child_of_or_equal_to(
        self,
        base_graph: OntologyGraph,
        sub: str,
        obj: str,
        expected: bool,
    ):
        assert (
            base_graph.is_child_of_or_equal_to(
                sub,
                obj,
            )
            is expected
        )

    @pytest.mark.parametrize(
        "sub, obj, expected",
        [
            ("HP:1", "HP:1", True),
            ("HP:01", "HP:1", True),
            ("HP:010", "HP:1", True),
        ],
    )
    def test_is_descendant_of_or_equal_to(
        self,
        base_graph: OntologyGraph,
        sub: str,
        obj: str,
        expected: bool,
    ):
        assert (
            base_graph.is_descendant_of_or_equal_to(
                sub,
                obj,
            )
            is expected
        )


@ddt.ddt
class TestCsrOntologyGraph(unittest.TestCase):
    """
    Tests of the `OntologyGraph` API that just happen to be using `BisectPoweredCsrOntologyGraph` as the implementation.
    """

    NODES: np.ndarray
    GRAPH: BisectPoweredCsrOntologyGraph

    @classmethod
    def setUpClass(cls) -> None:
        root = hpotk.TermId.from_curie("HP:1")
        curies = [
            "HP:01",
            "HP:010",
            "HP:011",
            "HP:0110",
            "HP:02",
            "HP:020",
            "HP:021",
            "HP:022",
            "HP:03",
            "HP:1",
        ]
        nodes = np.fromiter(map(hpotk.TermId.from_curie, curies), dtype=object)
        row = [0, 3, 5, 7, 9, 13, 14, 15, 16, 17, 20]
        col = [1, 2, 9, 0, 3, 0, 3, 1, 2, 5, 6, 7, 9, 4, 4, 4, 9, 0, 4, 8]
        data = [-1, -1, 1, 1, -1, 1, -1, 1, 1, -1, -1, -1, 1, 1, 1, 1, 1, -1, -1, -1]
        am = ImmutableCsrMatrix(row, col, data, shape=(len(nodes), len(nodes)), dtype=int)

        cls.NODES = nodes
        cls.GRAPH = BisectPoweredCsrOntologyGraph(root, nodes, am)

    def test_graph_queries_work_with_identified(self):
        # tests
        self.assertTrue(
            self.GRAPH.is_ancestor_of(
                SimpleIdentified.from_curie("HP:01"),
                SimpleIdentified.from_curie("HP:0110"),
            )
        )
        self.assertTrue(
            self.GRAPH.is_parent_of(
                SimpleIdentified.from_curie("HP:01"),
                SimpleIdentified.from_curie("HP:010"),
            )
        )
        self.assertTrue(
            self.GRAPH.is_child_of(
                SimpleIdentified.from_curie("HP:01"),
                SimpleIdentified.from_curie("HP:1"),
            )
        )
        self.assertTrue(
            self.GRAPH.is_descendant_of(
                SimpleIdentified.from_curie("HP:01"),
                SimpleIdentified.from_curie("HP:1"),
            )
        )
        self.assertTrue(self.GRAPH.is_leaf(SimpleIdentified.from_curie("HP:03")))

        # traversal
        self.assertSetEqual(
            set(self.GRAPH.get_ancestors(SimpleIdentified.from_curie("HP:010"))),
            {hpotk.TermId.from_curie("HP:01"), hpotk.TermId.from_curie("HP:1")},
        )
        self.assertSetEqual(
            set(self.GRAPH.get_parents(SimpleIdentified.from_curie("HP:01"))),
            {hpotk.TermId.from_curie("HP:1")},
        )
        self.assertSetEqual(
            set(self.GRAPH.get_children(SimpleIdentified.from_curie("HP:01"))),
            {hpotk.TermId.from_curie("HP:010"), hpotk.TermId.from_curie("HP:011")},
        )
        self.assertSetEqual(
            set(self.GRAPH.get_descendants(SimpleIdentified.from_curie("HP:01"))),
            {
                hpotk.TermId.from_curie("HP:010"),
                hpotk.TermId.from_curie("HP:011"),
                hpotk.TermId.from_curie("HP:0110"),
            },
        )

    def test_graph_queries_work_with_str(self):
        # tests
        self.assertTrue(self.GRAPH.is_ancestor_of("HP:01", "HP:0110"))
        self.assertTrue(self.GRAPH.is_parent_of("HP:01", "HP:010"))
        self.assertTrue(self.GRAPH.is_child_of("HP:01", "HP:1"))
        self.assertTrue(self.GRAPH.is_descendant_of("HP:01", "HP:1"))
        self.assertTrue(self.GRAPH.is_leaf("HP:03"))

        # traversal
        self.assertSetEqual(
            set(self.GRAPH.get_ancestors("HP:010")),
            {hpotk.TermId.from_curie("HP:01"), hpotk.TermId.from_curie("HP:1")},
        )
        self.assertSetEqual(set(self.GRAPH.get_parents("HP:01")), {hpotk.TermId.from_curie("HP:1")})
        self.assertSetEqual(
            set(self.GRAPH.get_children("HP:01")),
            {hpotk.TermId.from_curie("HP:010"), hpotk.TermId.from_curie("HP:011")},
        )
        self.assertSetEqual(
            set(self.GRAPH.get_descendants("HP:01")),
            {
                hpotk.TermId.from_curie("HP:010"),
                hpotk.TermId.from_curie("HP:011"),
                hpotk.TermId.from_curie("HP:0110"),
            },
        )

    @ddt.data(
        ("is_parent_of",),
        ("is_ancestor_of",),
        ("is_child_of",),
        ("is_descendant_of",),
    )
    @ddt.unpack
    def test_query_methods_with_unknown_source(self, func_name):
        existing = hpotk.TermId.from_curie("HP:1")
        unknown = hpotk.TermId.from_curie("HP:999")

        func = getattr(self.GRAPH, func_name)
        with self.assertRaises(ValueError) as ctx:
            func(existing, unknown)
        self.assertEqual("Term ID not found in the graph: HP:999", ctx.exception.args[0])

    def test_traversal_methods_produce_iterators(self):
        whatever = hpotk.TermId.from_curie("HP:1")
        self.assertIsInstance(self.GRAPH.get_parents(whatever), typing.Iterator)
        self.assertIsInstance(self.GRAPH.get_children(whatever), typing.Iterator)
        self.assertIsInstance(self.GRAPH.get_ancestors(whatever), typing.Iterator)
        self.assertIsInstance(self.GRAPH.get_descendants(whatever), typing.Iterator)


class SimpleIdentified(hpotk.model.Identified):
    @staticmethod
    def from_curie(curie: str):
        return SimpleIdentified(hpotk.TermId.from_curie(curie))

    def __init__(self, identifier: hpotk.TermId):
        self._id = identifier

    @property
    def identifier(self) -> hpotk.TermId:
        return self._id
